#  Copyright (c) ZenML GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Logger implementation."""

import builtins
import logging
import os
import re
import sys
from contextvars import ContextVar
from typing import TYPE_CHECKING, Any, Dict

if TYPE_CHECKING:
    from zenml.logging.step_logging import ArtifactStoreHandler

from rich.traceback import install as rich_tb_install

from zenml.constants import (
    ENABLE_RICH_TRACEBACK,
    ENV_ZENML_CAPTURE_PRINTS,
    ENV_ZENML_LOGGING_COLORS_DISABLED,
    ENV_ZENML_LOGGING_FORMAT,
    ENV_ZENML_SUPPRESS_LOGS,
    ZENML_LOGGING_VERBOSITY,
    ZENML_STORAGE_LOGGING_VERBOSITY,
    handle_bool_env_var,
)
from zenml.enums import LoggingLevels
from zenml.utils.context_utils import ContextVarList

ZENML_LOGGING_COLORS_DISABLED = handle_bool_env_var(
    ENV_ZENML_LOGGING_COLORS_DISABLED, False
)


step_names_in_console: ContextVar[bool] = ContextVar(
    "step_names_in_console", default=False
)
logging_handlers: ContextVarList["ArtifactStoreHandler"] = ContextVarList(
    "logging_handlers"
)


def _add_step_name_to_message(message: str) -> str:
    """Adds the step name to the message.

    Args:
        message: The message to add the step name to.

    Returns:
        The message with the step name added.
    """
    try:
        if step_names_in_console.get():
            from zenml.steps import get_step_context

            step_context = get_step_context()

            if step_context and message not in ["\n", ""]:
                # For progress bar updates (with \r), inject the step name after the \r
                if "\r" in message:
                    message = message.replace(
                        "\r", f"\r[{step_context.step_name}] "
                    )
                else:
                    message = f"[{step_context.step_name}] {message}"
    except Exception:
        # If we can't get step context, just use the original message
        pass

    return message


class CustomFormatter(logging.Formatter):
    """Formats logs according to custom specifications."""

    grey: str = "\x1b[90m"
    white: str = "\x1b[37m"
    pink: str = "\x1b[35m"
    green: str = "\x1b[32m"
    yellow: str = "\x1b[33m"
    red: str = "\x1b[31m"
    cyan: str = "\x1b[1;36m"
    bold_red: str = "\x1b[31;1m"
    purple: str = "\x1b[38;5;105m"
    blue: str = "\x1b[34m"
    reset: str = "\x1b[0m"

    def _get_format_template(self, record: logging.LogRecord) -> str:
        """Get the format template based on the logging level.

        Args:
            record: The log record to format.

        Returns:
            The format template string.
        """
        # Only include location info for DEBUG level
        if get_logging_level() == LoggingLevels.DEBUG:
            return "%(asctime)s - %(name)s - %(levelname)s - %(message)s (%(filename)s:%(lineno)d)"
        else:
            return "%(message)s"

    COLORS: Dict[LoggingLevels, str] = {
        LoggingLevels.DEBUG: grey,
        LoggingLevels.INFO: white,
        LoggingLevels.WARN: yellow,
        LoggingLevels.ERROR: red,
        LoggingLevels.CRITICAL: bold_red,
    }

    def format(self, record: logging.LogRecord) -> str:
        """Converts a log record to a (colored) string.

        Args:
            record: LogRecord generated by the code.

        Returns:
            A string formatted according to specifications.
        """
        # Get the template
        format_template = self._get_format_template(record)

        # Apply step name prepending if enabled (for console display)
        message = record.getMessage()
        try:
            if step_names_in_console.get():
                message = _add_step_name_to_message(message)
        except Exception:
            # If we can't get step context, just use the original message
            pass

        # Create a new record with the modified message
        modified_record = logging.LogRecord(
            name=record.name,
            level=record.levelno,
            pathname=record.pathname,
            lineno=record.lineno,
            msg=message,
            args=(),
            exc_info=record.exc_info,
        )

        if ZENML_LOGGING_COLORS_DISABLED:
            # If color formatting is disabled, use the default format without colors
            formatter = logging.Formatter(format_template)
            return formatter.format(modified_record)
        else:
            # Use color formatting
            log_fmt = (
                self.COLORS[LoggingLevels(record.levelno)]
                + format_template
                + self.reset
            )
            formatter = logging.Formatter(log_fmt)
            formatted_message = formatter.format(modified_record)
            quoted_groups = re.findall("`([^`]*)`", formatted_message)
            for quoted in quoted_groups:
                formatted_message = formatted_message.replace(
                    "`" + quoted + "`",
                    self.reset
                    + self.purple
                    + quoted
                    + self.COLORS.get(LoggingLevels(record.levelno)),
                )

            # Format URLs
            url_pattern = r"http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\\(\\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+"
            urls = re.findall(url_pattern, formatted_message)
            for url in urls:
                formatted_message = formatted_message.replace(
                    url,
                    self.reset
                    + self.blue
                    + url
                    + self.COLORS.get(LoggingLevels(record.levelno)),
                )
            return formatted_message


def get_logging_level() -> LoggingLevels:
    """Get logging level from the env variable.

    Returns:
        The logging level.

    Raises:
        KeyError: If the logging level is not found.
    """
    verbosity = ZENML_LOGGING_VERBOSITY.upper()
    if verbosity not in LoggingLevels.__members__:
        raise KeyError(
            f"Verbosity must be one of {list(LoggingLevels.__members__.keys())}"
        )
    return LoggingLevels[verbosity]


def get_storage_log_level() -> LoggingLevels:
    """Get storage logging level from the env variable with safe fallback.

    Returns:
        The storage logging level, defaulting to INFO if invalid.

    Raises:
        KeyError: If the storage logging level is not found.
    """
    verbosity = ZENML_STORAGE_LOGGING_VERBOSITY.upper()
    if verbosity not in LoggingLevels.__members__:
        raise KeyError(
            f"Verbosity must be one of {list(LoggingLevels.__members__.keys())}"
        )
    return LoggingLevels[verbosity]


def set_root_verbosity() -> None:
    """Set the root verbosity."""
    level = get_logging_level()
    if level != LoggingLevels.NOTSET:
        if ENABLE_RICH_TRACEBACK:
            rich_tb_install(show_locals=(level == LoggingLevels.DEBUG))

        logging.root.setLevel(level=level.value)
        get_logger(__name__).debug(
            f"Logging set to level: {logging.getLevelName(level.value)}"
        )
    else:
        logging.disable(sys.maxsize)
        logging.getLogger().disabled = True
        get_logger(__name__).debug("Logging NOTSET")


def wrapped_print(*args: Any, **kwargs: Any) -> None:
    """Wrapped print function.

    Args:
        *args: Arguments to print
        **kwargs: Keyword arguments for print
    """
    original_print = getattr(builtins, "_zenml_original_print")

    file_arg = kwargs.get("file", sys.stdout)

    # IMPORTANT: Don't intercept internal calls to any objects
    # other than sys.stdout and sys.stderr. This is especially
    # critical for handling tracebacks. The default logging
    # formatter uses StringIO to format tracebacks, we don't
    # want to intercept it and create a LogRecord about it.
    if file_arg not in (sys.stdout, sys.stderr):
        original_print(*args, **kwargs)

    # Convert print arguments to message
    message = " ".join(str(arg) for arg in args)

    # Call active handlers first (for storage)
    if message.strip():
        handlers = logging_handlers.get()

        for handler in handlers:
            try:
                # Create a LogRecord for the handler
                record = logging.LogRecord(
                    name="print",
                    level=logging.ERROR
                    if file_arg == sys.stderr
                    else logging.INFO,
                    pathname="",
                    lineno=0,
                    msg=message,
                    args=(),
                    exc_info=None,
                )
                # Check if handler's level would accept this record
                if record.levelno >= handler.level:
                    handler.emit(record)
            except Exception:
                # Don't let handler errors break print
                pass

    if step_names_in_console.get():
        message = _add_step_name_to_message(message)

    # Then call original print for console display
    original_print(message, *args[1:], **kwargs)


def setup_global_print_wrapping() -> None:
    """Set up global print() wrapping with context-aware handlers."""
    capture_prints = handle_bool_env_var(
        ENV_ZENML_CAPTURE_PRINTS, default=True
    )

    if not capture_prints or hasattr(__builtins__, "_zenml_original_print"):
        return

    # Store original and replace print
    setattr(builtins, "_zenml_original_print", builtins.print)
    setattr(builtins, "print", wrapped_print)


def get_formatter() -> logging.Formatter:
    """Get a configured logging formatter.

    Returns:
        The formatter.
    """
    if log_format := os.environ.get(ENV_ZENML_LOGGING_FORMAT, None):
        return logging.Formatter(fmt=log_format)
    else:
        return CustomFormatter()


def get_console_handler() -> Any:
    """Get console handler for logging.

    Returns:
        A console handler.
    """
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(get_formatter())
    # Set console handler level explicitly to console verbosity
    console_handler.setLevel(get_logging_level().value)
    return console_handler


def get_logger(logger_name: str) -> logging.Logger:
    """Main function to get logger name,.

    Args:
        logger_name: Name of logger to initialize.

    Returns:
        A logger object.
    """
    return logging.getLogger(logger_name)


def init_logging() -> None:
    """Initialize logging with default levels."""
    # Mute tensorflow cuda warnings
    os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"
    set_root_verbosity()

    # Check if console handler already exists to avoid duplicates
    root_logger = logging.getLogger()
    has_console_handler = any(
        isinstance(handler, logging.StreamHandler)
        and handler.stream == sys.stdout
        for handler in root_logger.handlers
    )

    # logging capture warnings
    logging.captureWarnings(True)

    if not has_console_handler:
        console_handler = logging.StreamHandler(sys.stdout)
        console_handler.setFormatter(get_formatter())
        # Set console handler level explicitly to console verbosity
        console_handler.setLevel(get_logging_level().value)
        root_logger.addHandler(console_handler)

    # Initialize global print wrapping
    setup_global_print_wrapping()

    # Enable logs if environment variable SUPPRESS_ZENML_LOGS is not set to True
    suppress_zenml_logs: bool = handle_bool_env_var(
        ENV_ZENML_SUPPRESS_LOGS, True
    )
    if suppress_zenml_logs:
        # suppress logger info messages
        suppressed_logger_names = [
            "urllib3",
            "azure.core.pipeline.policies.http_logging_policy",
            "grpc",
            "requests",
            "kfp",
            "tensorflow",
        ]
        for logger_name in suppressed_logger_names:
            logging.getLogger(logger_name).setLevel(logging.WARNING)

        # disable logger messages
        disabled_logger_names = [
            "rdbms_metadata_access_object",
            "backoff",
            "segment",
        ]
        for logger_name in disabled_logger_names:
            logging.getLogger(logger_name).setLevel(logging.WARNING)
            logging.getLogger(logger_name).disabled = True



"""Release [0.45.3].

Revision ID: 0.45.3
Revises: 729263e47b55
Create Date: 2023-10-18 14:39:09.972710

"""

# revision identifiers, used by Alembic.
revision = "0.45.3"
down_revision = "729263e47b55"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Dynamic pipeline runner."""

import contextvars
import copy
import inspect
import itertools
from concurrent.futures import ThreadPoolExecutor
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Union,
    overload,
)
from uuid import UUID

from zenml import ExternalArtifact
from zenml.artifacts.in_memory_cache import InMemoryArtifactCache
from zenml.client import Client
from zenml.config.compiler import Compiler
from zenml.enums import ExecutionMode, StepRuntime
from zenml.execution.pipeline.dynamic.outputs import (
    ArtifactFuture,
    MapResultsFuture,
    OutputArtifact,
    StepRunFuture,
    StepRunOutputs,
    StepRunOutputsFuture,
    _BaseStepRunFuture,
)
from zenml.execution.pipeline.dynamic.run_context import (
    DynamicPipelineRunContext,
)
from zenml.execution.pipeline.dynamic.utils import _Unmapped
from zenml.execution.step.utils import launch_step
from zenml.logger import get_logger
from zenml.logging.step_logging import setup_pipeline_logging
from zenml.models import (
    ArtifactVersionResponse,
    PipelineRunResponse,
    PipelineRunUpdate,
    PipelineSnapshotResponse,
)
from zenml.orchestrators.publish_utils import (
    publish_failed_pipeline_run,
    publish_successful_pipeline_run,
)
from zenml.pipelines.dynamic.pipeline_definition import DynamicPipeline
from zenml.pipelines.run_utils import create_placeholder_run
from zenml.stack import Stack
from zenml.steps.entrypoint_function_utils import StepArtifact
from zenml.steps.utils import OutputSignature
from zenml.utils import source_utils

if TYPE_CHECKING:
    from zenml.config import DockerSettings
    from zenml.config.step_configurations import Step, StepConfiguration
    from zenml.steps import BaseStep


logger = get_logger(__name__)


class DynamicPipelineRunner:
    """Dynamic pipeline runner."""

    def __init__(
        self,
        snapshot: "PipelineSnapshotResponse",
        run: Optional["PipelineRunResponse"],
    ) -> None:
        """Initialize the dynamic pipeline runner.

        Args:
            snapshot: The snapshot of the pipeline.
            run: The pipeline run.

        Raises:
            RuntimeError: If the snapshot has no associated stack.
        """
        if not snapshot.stack:
            raise RuntimeError("Missing stack for snapshot.")

        if (
            snapshot.pipeline_configuration.execution_mode
            != ExecutionMode.STOP_ON_FAILURE
        ):
            logger.warning(
                "Only the `%s` execution mode is supported for "
                "dynamic pipelines right now. "
                "The execution mode `%s` will be ignored.",
                ExecutionMode.STOP_ON_FAILURE,
                snapshot.pipeline_configuration.execution_mode,
            )

        self._snapshot = snapshot
        self._run = run
        # TODO: make this configurable
        self._executor = ThreadPoolExecutor(max_workers=10)
        self._pipeline: Optional["DynamicPipeline"] = None
        self._orchestrator = Stack.from_model(snapshot.stack).orchestrator
        self._orchestrator_run_id = (
            self._orchestrator.get_orchestrator_run_id()
        )
        self._futures: List[StepRunOutputsFuture] = []

    @property
    def pipeline(self) -> "DynamicPipeline":
        """The pipeline that the runner is executing.

        Raises:
            RuntimeError: If the pipeline can't be loaded.

        Returns:
            The pipeline that the runner is executing.
        """
        if self._pipeline is None:
            if (
                not self._snapshot.pipeline_spec
                or not self._snapshot.pipeline_spec.source
            ):
                raise RuntimeError("Missing pipeline source for snapshot.")

            pipeline = source_utils.load(self._snapshot.pipeline_spec.source)
            if not isinstance(pipeline, DynamicPipeline):
                raise RuntimeError(
                    "Invalid pipeline source: "
                    f"{self._snapshot.pipeline_spec.source.import_path}"
                )
            pipeline._configuration = self._snapshot.pipeline_configuration
            self._pipeline = pipeline

        return self._pipeline

    def run_pipeline(self) -> None:
        """Run the pipeline."""
        with setup_pipeline_logging(
            source="orchestrator",
            snapshot=self._snapshot,
        ) as logs_request:
            if self._run:
                run = Client().zen_store.update_run(
                    run_id=self._run.id,
                    run_update=PipelineRunUpdate(
                        orchestrator_run_id=self._orchestrator_run_id,
                        add_logs=[logs_request] if logs_request else None,
                    ),
                )
            else:
                run = create_placeholder_run(
                    snapshot=self._snapshot,
                    orchestrator_run_id=self._orchestrator_run_id,
                    logs=logs_request,
                )

            with InMemoryArtifactCache():
                with DynamicPipelineRunContext(
                    pipeline=self.pipeline,
                    run=run,
                    snapshot=self._snapshot,
                    runner=self,
                ):
                    self._orchestrator.run_init_hook(snapshot=self._snapshot)
                    try:
                        # TODO: step logging isn't threadsafe
                        # TODO: what should be allowed as pipeline returns?
                        #  (artifacts, json serializable, anything?)
                        #  how do we show it in the UI?
                        params = self.pipeline.configuration.parameters or {}
                        self.pipeline._call_entrypoint(**params)
                        # The pipeline function finished successfully, but some
                        # steps might still be running. We now wait for all of
                        # them and raise any exceptions that occurred.
                        self.await_all_step_run_futures()
                    except:
                        publish_failed_pipeline_run(run.id)
                        logger.error(
                            "Pipeline run failed. All in-progress step runs "
                            "will still finish executing."
                        )
                        raise
                    finally:
                        self._orchestrator.run_cleanup_hook(
                            snapshot=self._snapshot
                        )
                        self._executor.shutdown(wait=True, cancel_futures=True)

                    publish_successful_pipeline_run(run.id)

    @overload
    def launch_step(
        self,
        step: "BaseStep",
        id: Optional[str],
        args: Tuple[Any, ...],
        kwargs: Dict[str, Any],
        after: Union["StepRunFuture", Sequence["StepRunFuture"], None] = None,
        concurrent: Literal[False] = False,
    ) -> StepRunOutputs: ...

    @overload
    def launch_step(
        self,
        step: "BaseStep",
        id: Optional[str],
        args: Tuple[Any, ...],
        kwargs: Dict[str, Any],
        after: Union["StepRunFuture", Sequence["StepRunFuture"], None] = None,
        concurrent: Literal[True] = True,
    ) -> "StepRunOutputsFuture": ...

    def launch_step(
        self,
        step: "BaseStep",
        id: Optional[str],
        args: Tuple[Any, ...],
        kwargs: Dict[str, Any],
        after: Union["StepRunFuture", Sequence["StepRunFuture"], None] = None,
        concurrent: bool = False,
    ) -> Union[StepRunOutputs, "StepRunOutputsFuture"]:
        """Launch a step.

        Args:
            step: The step to launch.
            id: The invocation ID of the step.
            args: The arguments for the step function.
            kwargs: The keyword arguments for the step function.
            after: The step run output futures to wait for.
            concurrent: Whether to launch the step concurrently.

        Returns:
            The step run outputs or a future for the step run outputs.
        """
        step = step.copy()
        compiled_step = compile_dynamic_step_invocation(
            snapshot=self._snapshot,
            pipeline=self.pipeline,
            step=step,
            id=id,
            inputs=convert_to_keyword_arguments(step.entrypoint, args, kwargs),
            after=after,
        )

        def _launch() -> StepRunOutputs:
            step_run = launch_step(
                snapshot=self._snapshot,
                step=compiled_step,
                orchestrator_run_id=self._orchestrator_run_id,
                retry=_should_retry_locally(
                    compiled_step,
                    self._snapshot.pipeline_configuration.docker_settings,
                ),
            )
            return _load_step_run_outputs(step_run.id)

        if concurrent:
            ctx = contextvars.copy_context()
            future = self._executor.submit(ctx.run, _launch)
            step_run_future = StepRunOutputsFuture(
                wrapped=future,
                invocation_id=compiled_step.spec.invocation_id,
                output_keys=list(compiled_step.config.outputs),
            )
            self._futures.append(step_run_future)
            return step_run_future
        else:
            return _launch()

    def map(
        self,
        step: "BaseStep",
        args: Tuple[Any, ...],
        kwargs: Dict[str, Any],
        after: Union["StepRunFuture", Sequence["StepRunFuture"], None] = None,
        product: bool = False,
    ) -> "MapResultsFuture":
        """Map over step inputs.

        Args:
            step: The step to run.
            args: The arguments for the step function.
            kwargs: The keyword arguments for the step function.
            after: The step run output futures to wait for before executing the
                steps.
            product: Whether to produce a cartesian product of the mapped
                inputs.

        Returns:
            A future that represents the map results.
        """
        kwargs = convert_to_keyword_arguments(step.entrypoint, args, kwargs)
        kwargs = await_step_inputs(kwargs)
        step_inputs = expand_mapped_inputs(kwargs, product=product)

        step_run_futures = [
            self.launch_step(
                step,
                id=None,
                args=(),
                kwargs=inputs,
                after=after,
                concurrent=True,
            )
            for inputs in step_inputs
        ]

        return MapResultsFuture(futures=step_run_futures)

    def await_all_step_run_futures(self) -> None:
        """Await all step run output futures."""
        for future in self._futures:
            future._wait()
        self._futures = []


def compile_dynamic_step_invocation(
    snapshot: "PipelineSnapshotResponse",
    pipeline: "DynamicPipeline",
    step: "BaseStep",
    inputs: Dict[str, Any],
    after: Union[
        "StepRunFuture",
        "ArtifactFuture",
        Sequence[Union["StepRunFuture", "ArtifactFuture"]],
        None,
    ] = None,
    id: Optional[str] = None,
) -> "Step":
    """Compile a dynamic step invocation.

    Args:
        snapshot: The snapshot.
        pipeline: The dynamic pipeline.
        step: The step to compile.
        id: Custom invocation ID.
        inputs: The inputs for the step function.
        after: The step run output futures to wait for.

    Returns:
        The compiled step.
    """
    upstream_steps = set()

    if isinstance(after, _BaseStepRunFuture):
        after._wait()
        upstream_steps.add(after.invocation_id)
    elif isinstance(after, MapResultsFuture):
        for future in after:
            future._wait()
            upstream_steps.add(future.invocation_id)
    elif isinstance(after, Sequence):
        for item in after:
            if isinstance(item, _BaseStepRunFuture):
                item._wait()
                upstream_steps.add(item.invocation_id)
            elif isinstance(item, MapResultsFuture):
                for future in item:
                    future._wait()
                    upstream_steps.add(future.invocation_id)

    inputs = await_step_inputs(inputs)

    for value in inputs.values():
        if isinstance(value, OutputArtifact):
            upstream_steps.add(value.step_name)

        if (
            isinstance(value, Sequence)
            and value
            and all(isinstance(item, OutputArtifact) for item in value)
        ):
            upstream_steps.update(item.step_name for item in value)

    default_parameters = {
        key: value
        for key, value in convert_to_keyword_arguments(
            step.entrypoint, (), inputs, apply_defaults=True
        ).items()
        if key not in inputs
    }

    input_artifacts = {}
    external_artifacts = {}
    for name, value in inputs.items():
        if isinstance(value, OutputArtifact):
            input_artifacts[name] = [
                StepArtifact(
                    invocation_id=value.step_name,
                    output_name=value.output_name,
                    annotation=OutputSignature(resolved_annotation=Any),
                    pipeline=pipeline,
                    chunk_index=value.chunk_index,
                    chunk_size=value.chunk_size,
                )
            ]
        elif (
            isinstance(value, list)
            and value
            and all(isinstance(item, OutputArtifact) for item in value)
        ):
            input_artifacts[name] = [
                StepArtifact(
                    invocation_id=item.step_name,
                    output_name=item.output_name,
                    annotation=OutputSignature(resolved_annotation=Any),
                    pipeline=pipeline,
                    chunk_index=item.chunk_index,
                    chunk_size=item.chunk_size,
                )
                for item in value
            ]
        elif isinstance(value, (ArtifactVersionResponse, ExternalArtifact)):
            external_artifacts[name] = value
        else:
            # TODO: should some of these be parameters?
            external_artifacts[name] = ExternalArtifact(value=value)

    if template := get_config_template(snapshot, step, pipeline):
        step._configuration = template.config.model_copy(
            update={"template": template.spec.invocation_id}
        )

    invocation_id = pipeline.add_step_invocation(
        step=step,
        custom_id=id,
        allow_id_suffix=not id,
        input_artifacts=input_artifacts,
        external_artifacts=external_artifacts,
        upstream_steps=upstream_steps,
        default_parameters=default_parameters,
        parameters={},
        model_artifacts_or_metadata={},
        client_lazy_loaders={},
    )

    return Compiler()._compile_step_invocation(
        invocation=pipeline.invocations[invocation_id],
        stack=Client().active_stack,
        step_config=None,
        pipeline=pipeline,
    )


def _load_step_run_outputs(step_run_id: UUID) -> StepRunOutputs:
    """Load the outputs of a step run.

    Args:
        step_run_id: The ID of the step run.

    Returns:
        The outputs of the step run.
    """
    step_run = Client().zen_store.get_run_step(step_run_id)

    def _convert_output_artifact(
        output_name: str, artifact: ArtifactVersionResponse
    ) -> OutputArtifact:
        return OutputArtifact(
            output_name=output_name,
            step_name=step_run.name,
            **artifact.model_dump(),
        )

    output_artifacts = step_run.regular_outputs
    if len(output_artifacts) == 0:
        return None
    elif len(output_artifacts) == 1:
        name, artifact = next(iter(output_artifacts.items()))
        return _convert_output_artifact(output_name=name, artifact=artifact)
    else:
        # Make sure we return them in the same order as they're defined in the
        # step configuration, as we don't enforce any ordering in the DB.
        return tuple(
            _convert_output_artifact(
                output_name=name, artifact=output_artifacts[name]
            )
            for name in step_run.config.outputs.keys()
        )


def _should_retry_locally(
    step: "Step", pipeline_docker_settings: "DockerSettings"
) -> bool:
    """Determine if a step should be retried locally.

    Args:
        step: The step.
        pipeline_docker_settings: The Docker settings of the parent pipeline.

    Returns:
        Whether the step should be retried locally.
    """
    if step.config.step_operator:
        return True

    runtime = get_step_runtime(
        step_config=step.config,
        pipeline_docker_settings=pipeline_docker_settings,
    )
    if runtime == StepRuntime.INLINE:
        return True
    else:
        # Running in isolated mode with the orchestrator
        return (
            not Client().active_stack.orchestrator.config.handles_step_retries
        )


def get_step_runtime(
    step_config: "StepConfiguration",
    pipeline_docker_settings: "DockerSettings",
) -> StepRuntime:
    """Determine if a step should be run in process.

    Args:
        step_config: The step configuration.
        pipeline_docker_settings: The Docker settings of the parent pipeline.

    Returns:
        The runtime for the step.
    """
    if step_config.step_operator:
        return StepRuntime.ISOLATED

    if not Client().active_stack.orchestrator.can_run_isolated_steps:
        return StepRuntime.INLINE

    runtime = step_config.runtime

    if runtime is None:
        if not step_config.resource_settings.empty:
            runtime = StepRuntime.ISOLATED
        elif step_config.docker_settings != pipeline_docker_settings:
            runtime = StepRuntime.ISOLATED
        else:
            runtime = StepRuntime.INLINE

    return runtime


def get_config_template(
    snapshot: "PipelineSnapshotResponse",
    step: "BaseStep",
    pipeline: "DynamicPipeline",
) -> Optional["Step"]:
    """Get the config template for a step executed in a dynamic pipeline.

    Args:
        snapshot: The snapshot of the pipeline.
        step: The step to get the config template for.
        pipeline: The dynamic pipeline that the step is being executed in.

    Returns:
        The config template for the step.
    """
    for index, step_ in enumerate(pipeline.depends_on):
        if step_._static_id == step._static_id:
            break
    else:
        return None

    return list(snapshot.step_configurations.values())[index]


def expand_mapped_inputs(
    inputs: Dict[str, Any],
    product: bool = False,
) -> List[Dict[str, Any]]:
    """Find the mapped and unmapped inputs of a step.

    Args:
        inputs: The step function inputs.
        product: Whether to produce a cartesian product of the mapped inputs.

    Raises:
        RuntimeError: If no mapped inputs are found or the input combinations
            are not valid.

    Returns:
        The step inputs.
    """
    static_inputs: Dict[str, Any] = {}
    mapped_input_names: List[str] = []
    mapped_inputs: List[Tuple[OutputArtifact, ...]] = []

    for key, value in inputs.items():
        if isinstance(value, _Unmapped):
            static_inputs[key] = value.value
        elif isinstance(value, OutputArtifact):
            if value.item_count is None:
                static_inputs[key] = value
            elif value.item_count == 0:
                raise RuntimeError(
                    f"Artifact `{value.id}` has 0 items and cannot be mapped "
                    "over. Wrap it with the `unmapped(...)` function to pass "
                    "the artifact without mapping over it."
                )
            else:
                mapped_input_names.append(key)
                mapped_inputs.append(
                    tuple(
                        value.model_copy(
                            update={"chunk_index": i, "chunk_size": 1}
                        )
                        for i in range(value.item_count)
                    )
                )
        elif (
            isinstance(value, ArtifactVersionResponse)
            and value.item_count is not None
        ):
            static_inputs[key] = value
            logger.warning(
                "Received sequence-like artifact for step input `%s`. Mapping "
                "over artifacts that are not step output artifacts is "
                "currently not supported, and the complete artifact will be "
                "passed to all steps. If you want to silence this warning, "
                "wrap your input with the `unmapped(...)` function.",
                key,
            )
        elif isinstance(value, Sequence):
            logger.warning(
                "Received sequence-like data for step input `%s`. Mapping over "
                "data that is not a step output artifact is currently not "
                "supported, and the complete data will be passed to all steps. "
                "If you want to silence this warning, wrap your input with the "
                "`unmapped(...)` function.",
                key,
            )
            static_inputs[key] = value
        else:
            static_inputs[key] = value

    if len(mapped_inputs) == 0:
        raise RuntimeError(
            "No inputs to map over found. When calling `.map(...)` or "
            "`.product(...)` on a step, you need to pass at least one "
            "sequence-like step output of a previous step as input."
        )

    step_inputs = []

    if product:
        for input_combination in itertools.product(*mapped_inputs):
            all_inputs = copy.deepcopy(static_inputs)
            for name, value in zip(mapped_input_names, input_combination):
                all_inputs[name] = value
            step_inputs.append(all_inputs)
    else:
        item_counts = [len(inputs) for inputs in mapped_inputs]
        if not all(count == item_counts[0] for count in item_counts):
            raise RuntimeError(
                f"All mapped input artifacts must have the same "
                "item counts, but you passed artifacts with item counts "
                f"{item_counts}. If you want "
                "to pass sequence-like artifacts without mapping over "
                "them, wrap them with the `unmapped(...)` function."
            )

        for i in range(item_counts[0]):
            all_inputs = copy.deepcopy(static_inputs)
            for name, artifact in zip(
                mapped_input_names,
                [artifact_list[i] for artifact_list in mapped_inputs],
            ):
                all_inputs[name] = artifact
            step_inputs.append(all_inputs)

    return step_inputs


def convert_to_keyword_arguments(
    func: Callable[..., Any],
    args: Tuple[Any, ...],
    kwargs: Dict[str, Any],
    apply_defaults: bool = False,
) -> Dict[str, Any]:
    """Convert function arguments to keyword arguments.

    Args:
        func: The function to convert the arguments to keyword arguments for.
        args: The arguments to convert to keyword arguments.
        kwargs: The keyword arguments to convert to keyword arguments.
        apply_defaults: Whether to apply the function default values.

    Returns:
        The keyword arguments.
    """
    signature = inspect.signature(func, follow_wrapped=True)
    bound_args = signature.bind_partial(*args, **kwargs)
    if apply_defaults:
        bound_args.apply_defaults()

    return bound_args.arguments


def await_step_inputs(inputs: Dict[str, Any]) -> Dict[str, Any]:
    """Await the inputs of a step.

    Args:
        inputs: The inputs of the step.

    Raises:
        RuntimeError: If a step run future with multiple output artifacts is
            passed as an input.

    Returns:
        The awaited inputs.
    """
    result = {}
    for key, value in inputs.items():
        if isinstance(value, MapResultsFuture):
            value = value.futures

        if (
            isinstance(value, Sequence)
            and value
            and all(isinstance(item, StepRunOutputsFuture) for item in value)
        ):
            if any(len(item._output_keys) != 1 for item in value):
                raise RuntimeError(
                    f"Invalid step input `{key}`: Passing a future that refers "
                    "to multiple output artifacts as an input to another step "
                    "is not allowed."
                )
            value = [item.artifacts() for item in value]
        elif isinstance(value, StepRunOutputsFuture):
            if len(value._output_keys) != 1:
                raise RuntimeError(
                    f"Invalid step input `{key}`: Passing a future that refers "
                    "to multiple output artifacts as an input to another step "
                    "is not allowed."
                )
            value = value.artifacts()

        if (
            isinstance(value, Sequence)
            and value
            and all(isinstance(item, ArtifactFuture) for item in value)
        ):
            value = [item.result() for item in value]

        if isinstance(value, ArtifactFuture):
            value = value.result()

        result[key] = value

    return result



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""ZenML Step HeartBeat functionality."""

import _thread
import logging
import threading
import time
from uuid import UUID

from zenml.enums import ExecutionStatus

logger = logging.getLogger(__name__)


class StepHeartBeatTerminationException(Exception):
    """Custom exception class for heartbeat termination."""

    pass


class StepHeartbeatWorker:
    """Worker class implementing heartbeat polling and remote termination."""

    STEP_HEARTBEAT_INTERVAL_SECONDS = 60

    def __init__(self, step_id: UUID):
        """Heartbeat worker constructor.

        Args:
            step_id: The step id heartbeat is running for.
        """
        self._step_id = step_id

        self._thread: threading.Thread | None = None
        self._running: bool = False
        self._terminated: bool = False

    # properties

    @property
    def is_running(self) -> bool:
        """Property function for running signal.

        Returns:
            True if the worker is running.
        """
        return self._running

    @property
    def is_terminated(self) -> bool:
        """Property function for termination signal.

        Returns:
            True if the worker has been terminated.
        """
        return self._terminated

    @property
    def interval(self) -> int:
        """Property function for heartbeat interval.

        Returns:
            The heartbeat polling interval value.
        """
        return self.STEP_HEARTBEAT_INTERVAL_SECONDS

    @property
    def name(self) -> str:
        """Property function for heartbeat worker name.

        Returns:
            The name of the heartbeat worker.
        """
        return f"HeartBeatWorker-{self.step_id}"

    @property
    def step_id(self) -> UUID:
        """Property function for heartbeat worker step ID.

        Returns:
            The id of the step heartbeat is running for.
        """
        return self._step_id

    # public functions

    def start(self) -> None:
        """Start the heartbeat worker on a background thread."""
        if self._thread and self._thread.is_alive():
            return

        self._running = True
        self._terminated = False
        self._thread = threading.Thread(
            target=self._run, name=self.name, daemon=True
        )
        self._thread.start()
        logger.debug(
            "Daemon thread %s started (interval=%s)", self.name, self.interval
        )

    def stop(self) -> None:
        """Stops the heartbeat worker."""
        if not self._running:
            return
        self._running = False
        logger.debug("%s stop requested", self.name)

    def is_alive(self) -> bool:
        """Liveness of the heartbeat worker thread.

        Returns:
            True if the heartbeat worker thread is alive, False otherwise.
        """
        t = self._thread
        return bool(t and t.is_alive())

    def _run(self) -> None:
        logger.debug("%s run() loop entered", self.name)
        try:
            while self._running:
                try:
                    self._heartbeat()
                    # One-shot: signal the main thread and stop the loop.
                    if self._terminated:
                        logger.info(
                            "%s is remotely stopped, interrupting main thread",
                            self.name,
                        )
                        _thread.interrupt_main()  # raises KeyboardInterrupt in main thread
                        self._running = False
                    # Ensure we stop our own loop as well.

                except Exception as exc:
                    # Log-and-continue policy for all other errors.
                    logger.debug(
                        "%s heartbeat() failed with %s", self.name, str(exc)
                    )
                # Sleep after each attempt (even after errors, unless stopped).
                if self._running:
                    time.sleep(self.interval)
        finally:
            logger.debug("%s run() loop exiting", self.name)

    def _heartbeat(self) -> None:
        from zenml.config.global_config import GlobalConfiguration

        store = GlobalConfiguration().zen_store

        response = store.update_step_heartbeat(step_run_id=self.step_id)

        if response.status in {
            ExecutionStatus.STOPPED,
            ExecutionStatus.STOPPING,
        }:
            self._terminated = True



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Entrypoint of the Kubernetes master/orchestrator pod."""

import argparse
import random
import socket
import threading
import time
from typing import List, Optional, Tuple, cast
from uuid import UUID

from kubernetes import client as k8s_client
from kubernetes.client.rest import ApiException

from zenml.client import Client
from zenml.entrypoints.step_entrypoint_configuration import (
    StepEntrypointConfiguration,
)
from zenml.enums import ExecutionMode, ExecutionStatus, MetadataResourceTypes
from zenml.exceptions import AuthorizationException
from zenml.integrations.kubernetes import kube_utils
from zenml.integrations.kubernetes.constants import (
    ENV_ZENML_KUBERNETES_RUN_ID,
    KUBERNETES_SECRET_TOKEN_KEY_NAME,
    ORCHESTRATOR_RUN_ID_ANNOTATION_KEY,
    RUN_ID_ANNOTATION_KEY,
    STEP_NAME_ANNOTATION_KEY,
)
from zenml.integrations.kubernetes.flavors.kubernetes_orchestrator_flavor import (
    KubernetesOrchestratorSettings,
)
from zenml.integrations.kubernetes.manifest_utils import (
    build_job_manifest,
    build_pod_manifest,
    pod_template_manifest_from_pod,
)
from zenml.integrations.kubernetes.orchestrators.dag_runner import (
    DagRunner,
    InterruptMode,
    Node,
    NodeStatus,
)
from zenml.integrations.kubernetes.orchestrators.kubernetes_orchestrator import (
    KubernetesOrchestrator,
)
from zenml.logger import get_logger
from zenml.logging.step_logging import setup_orchestrator_logging
from zenml.models import (
    PipelineRunResponse,
    PipelineRunUpdate,
    PipelineSnapshotResponse,
    RunMetadataResource,
)
from zenml.orchestrators import publish_utils
from zenml.orchestrators.step_run_utils import (
    StepRunRequestFactory,
    fetch_step_runs_by_names,
    publish_cached_step_run,
)
from zenml.orchestrators.utils import (
    get_config_environment_vars,
)
from zenml.pipelines.run_utils import create_placeholder_run
from zenml.utils import env_utils

logger = get_logger(__name__)


def parse_args() -> argparse.Namespace:
    """Parse entrypoint arguments.

    Returns:
        Parsed args.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("--snapshot_id", type=str, required=True)
    parser.add_argument("--run_id", type=str, required=False)
    return parser.parse_args()


def _get_orchestrator_job_state(
    batch_api: k8s_client.BatchV1Api, namespace: str, job_name: str
) -> Tuple[Optional[UUID], Optional[str]]:
    """Get the existing status of the orchestrator job.

    Args:
        batch_api: The batch api.
        namespace: The namespace.
        job_name: The name of the orchestrator job.

    Returns:
        The run id and orchestrator run id.
    """
    run_id = None
    orchestrator_run_id = None

    job = kube_utils.get_job(
        batch_api=batch_api,
        namespace=namespace,
        job_name=job_name,
    )

    if job.metadata and job.metadata.annotations:
        annotations = job.metadata.annotations

        run_id = annotations.get(RUN_ID_ANNOTATION_KEY, None)
        orchestrator_run_id = annotations.get(
            ORCHESTRATOR_RUN_ID_ANNOTATION_KEY, None
        )

    return UUID(run_id) if run_id else None, orchestrator_run_id


def _reconstruct_nodes(
    snapshot: PipelineSnapshotResponse,
    pipeline_run: PipelineRunResponse,
    namespace: str,
    batch_api: k8s_client.BatchV1Api,
) -> List[Node]:
    """Reconstruct the nodes from the pipeline run.

    Args:
        snapshot: The snapshot.
        pipeline_run: The pipeline run.
        namespace: The namespace.
        batch_api: The batch api.

    Returns:
        The reconstructed nodes.
    """
    nodes = {
        step_name: Node(id=step_name, upstream_nodes=step.spec.upstream_steps)
        for step_name, step in snapshot.step_configurations.items()
    }

    for step_name, existing_step_run in pipeline_run.steps.items():
        node = nodes[step_name]
        if existing_step_run.status.is_successful:
            node.status = NodeStatus.COMPLETED
        elif existing_step_run.status.is_finished:
            node.status = NodeStatus.FAILED

    job_list = kube_utils.list_jobs(
        batch_api=batch_api,
        namespace=namespace,
        label_selector=f"run_id={pipeline_run.id}",
    )
    for job in job_list.items:
        annotations = job.metadata.annotations or {}
        if step_name_annotation := annotations.get(
            STEP_NAME_ANNOTATION_KEY, None
        ):
            node = nodes[str(step_name_annotation)]
            node.metadata["job_name"] = job.metadata.name

            if node.status == NodeStatus.NOT_READY:
                # The step is not finished in the ZenML database, so we base it
                # on the job status.
                node_status = NodeStatus.RUNNING
                if job.status.conditions:
                    for condition in job.status.conditions:
                        if (
                            condition.type == "Complete"
                            and condition.status == "True"
                        ):
                            node_status = NodeStatus.COMPLETED
                            break
                        elif (
                            condition.type == "Failed"
                            and condition.status == "True"
                        ):
                            node_status = NodeStatus.FAILED
                            break

                node.status = node_status
                logger.debug(
                    "Existing job for step `%s` status: %s.",
                    step_name,
                    node_status,
                )

    return list(nodes.values())


def main() -> None:
    """Entrypoint of the k8s master/orchestrator pod.

    Raises:
        RuntimeError: If the orchestrator pod is not associated with a job.
    """
    logger.info("Orchestrator pod started.")

    args = parse_args()

    orchestrator_pod_name = socket.gethostname()

    client = Client()
    snapshot = client.get_snapshot(args.snapshot_id)
    active_stack = client.active_stack
    orchestrator = active_stack.orchestrator
    assert isinstance(orchestrator, KubernetesOrchestrator)
    namespace = orchestrator.config.kubernetes_namespace

    pipeline_settings = cast(
        KubernetesOrchestratorSettings,
        orchestrator.get_settings(snapshot),
    )

    # Get a Kubernetes client from the active Kubernetes orchestrator, but
    # override the `incluster` setting to `True` since we are running inside
    # the Kubernetes cluster.
    api_client_config = orchestrator.get_kube_client(
        incluster=True
    ).configuration
    api_client_config.connection_pool_maxsize = (
        pipeline_settings.max_parallelism
    )
    kube_client = k8s_client.ApiClient(api_client_config)
    core_api = k8s_client.CoreV1Api(kube_client)
    batch_api = k8s_client.BatchV1Api(kube_client)

    job_name = kube_utils.get_parent_job_name(
        core_api=core_api,
        pod_name=orchestrator_pod_name,
        namespace=namespace,
    )
    if not job_name:
        raise RuntimeError("Failed to fetch job name for orchestrator pod.")

    run_id, orchestrator_run_id = _get_orchestrator_job_state(
        batch_api=batch_api,
        namespace=namespace,
        job_name=job_name,
    )
    existing_logs_response = None

    if run_id and orchestrator_run_id:
        logger.info("Continuing existing run `%s`.", run_id)
        pipeline_run = client.get_pipeline_run(run_id)
        nodes = _reconstruct_nodes(
            snapshot=snapshot,
            pipeline_run=pipeline_run,
            namespace=namespace,
            batch_api=batch_api,
        )
        logger.debug("Reconstructed nodes: %s", nodes)

        # Continue logging to the same log file if it exists
        for log_response in pipeline_run.log_collection or []:
            if log_response.source == "orchestrator":
                existing_logs_response = log_response
                break
    else:
        orchestrator_run_id = orchestrator_pod_name
        if args.run_id:
            pipeline_run = client.zen_store.update_run(
                run_id=args.run_id,
                run_update=PipelineRunUpdate(
                    orchestrator_run_id=orchestrator_run_id
                ),
            )
        else:
            pipeline_run = create_placeholder_run(
                snapshot=snapshot,
                orchestrator_run_id=orchestrator_run_id,
            )

        # Store in the job annotations so we can continue the run if the pod
        # is restarted
        kube_utils.update_job(
            batch_api=batch_api,
            namespace=namespace,
            job_name=job_name,
            annotations={
                RUN_ID_ANNOTATION_KEY: str(pipeline_run.id),
                ORCHESTRATOR_RUN_ID_ANNOTATION_KEY: orchestrator_run_id,
            },
        )
        nodes = [
            Node(id=step_name, upstream_nodes=step.spec.upstream_steps)
            for step_name, step in snapshot.step_configurations.items()
        ]

    logs_context = setup_orchestrator_logging(
        run_id=pipeline_run.id,
        snapshot=snapshot,
        logs_response=existing_logs_response,
    )

    with logs_context:
        step_command = StepEntrypointConfiguration.get_entrypoint_command()
        mount_local_stores = active_stack.orchestrator.config.is_local

        shared_env, secrets = get_config_environment_vars()
        shared_env[ENV_ZENML_KUBERNETES_RUN_ID] = orchestrator_run_id

        owner_references = None
        if not orchestrator.config.skip_owner_references:
            try:
                owner_references = kube_utils.get_pod_owner_references(
                    core_api=core_api,
                    pod_name=orchestrator_pod_name,
                    namespace=namespace,
                )
                # Make sure None of the owner references are marked as
                # controllers of the created pod, which messes with the
                # garbage collection logic.
                for owner_reference in owner_references:
                    owner_reference.controller = False
            except Exception as e:
                logger.warning(f"Failed to get pod owner references: {str(e)}")

        step_run_request_factory = StepRunRequestFactory(
            snapshot=snapshot,
            pipeline_run=pipeline_run,
            stack=active_stack,
        )
        step_runs = {}

        base_labels = {
            "project_id": kube_utils.sanitize_label(str(snapshot.project_id)),
            "run_id": kube_utils.sanitize_label(str(pipeline_run.id)),
            "run_name": kube_utils.sanitize_label(str(pipeline_run.name)),
            "pipeline": kube_utils.sanitize_label(
                snapshot.pipeline_configuration.name
            ),
        }

        def _cache_step_run_if_possible(step_name: str) -> bool:
            if not step_run_request_factory.has_caching_enabled(step_name):
                return False

            step_run_request = step_run_request_factory.create_request(
                step_name
            )
            try:
                step_run_request_factory.populate_request(step_run_request)
            except Exception as e:
                logger.error(
                    f"Failed to populate step run request for step {step_name}: {e}"
                )
                return False

            if step_run_request.status == ExecutionStatus.CACHED:
                step_run = publish_cached_step_run(
                    step_run_request, pipeline_run
                )
                step_runs[step_name] = step_run
                logger.info("Using cached version of step `%s`.", step_name)
                return True

            return False

        startup_lock = threading.Lock()
        last_startup_time: float = 0.0

        def start_step_job(node: Node) -> NodeStatus:
            """Run a pipeline step in a separate Kubernetes pod.

            Args:
                node: The node to start.

            Returns:
                The status of the node.
            """
            step_name = node.id
            step_config = snapshot.step_configurations[step_name].config
            settings = step_config.settings.get(
                "orchestrator.kubernetes", None
            )
            settings = KubernetesOrchestratorSettings.model_validate(
                settings.model_dump() if settings else {}
            )
            if not pipeline_settings.prevent_orchestrator_pod_caching:
                if _cache_step_run_if_possible(step_name):
                    return NodeStatus.COMPLETED

            step_labels = base_labels.copy()
            step_labels["step_name"] = kube_utils.sanitize_label(step_name)
            step_annotations = {
                STEP_NAME_ANNOTATION_KEY: step_name,
            }

            step_env = shared_env.copy()
            step_env.update(
                env_utils.get_step_environment(
                    step_config=step_config, stack=active_stack
                )
            )

            image = KubernetesOrchestrator.get_image(
                snapshot=snapshot, step_name=step_name
            )
            step_args = StepEntrypointConfiguration.get_entrypoint_arguments(
                step_name=step_name, snapshot_id=snapshot.id
            )

            # We set some default minimum memory resource requests for the step pod
            # here if the user has not specified any, because the step pod takes up
            # some memory resources itself and, if not specified, the pod will be
            # scheduled on any node regardless of available memory and risk
            # negatively impacting or even crashing the node due to memory pressure.
            pod_settings = kube_utils.apply_default_resource_requests(
                memory="400Mi",
                pod_settings=settings.pod_settings,
            )

            if orchestrator.config.pass_zenml_token_as_secret:
                step_env.pop("ZENML_STORE_API_TOKEN", None)
                secret_name = orchestrator.get_token_secret_name(snapshot.id)
                pod_settings.env.append(
                    {
                        "name": "ZENML_STORE_API_TOKEN",
                        "valueFrom": {
                            "secretKeyRef": {
                                "name": secret_name,
                                "key": KUBERNETES_SECRET_TOKEN_KEY_NAME,
                            }
                        },
                    }
                )
            else:
                step_env.update(secrets)

            pod_manifest = build_pod_manifest(
                pod_name=None,
                image_name=image,
                command=step_command,
                args=step_args,
                env=step_env,
                privileged=settings.privileged,
                pod_settings=pod_settings,
                service_account_name=settings.step_pod_service_account_name
                or settings.service_account_name,
                mount_local_stores=mount_local_stores,
                termination_grace_period_seconds=settings.pod_stop_grace_period,
                labels=step_labels,
            )

            retry_config = step_config.retry
            backoff_limit = (
                retry_config.max_retries if retry_config else 0
            ) + settings.backoff_limit_margin

            pod_failure_policy = settings.pod_failure_policy or {
                # These rules are applied sequentially. This means any failure in
                # the main container will count towards the max retries. Any other
                # disruption will not count towards the max retries.
                "rules": [
                    # If the main container fails, we count it towards the max
                    # retries.
                    {
                        "action": "Count",
                        "onExitCodes": {
                            "containerName": "main",
                            "operator": "NotIn",
                            "values": [0],
                        },
                    },
                    # If the pod is interrupted at any other time, we don't count
                    # it as a retry
                    {
                        "action": "Ignore",
                        "onPodConditions": [
                            {
                                "type": "DisruptionTarget",
                            }
                        ],
                    },
                ]
            }

            job_name = settings.job_name_prefix or ""
            random_prefix = "".join(random.choices("0123456789abcdef", k=8))
            job_name += f"-{random_prefix}-{step_name}-{snapshot.pipeline_configuration.name}"
            # The job name will be used as a label on the pods, so we need to make
            # sure it doesn't exceed the label length limit
            job_name = kube_utils.sanitize_label(job_name)

            job_manifest = build_job_manifest(
                job_name=job_name,
                pod_template=pod_template_manifest_from_pod(pod_manifest),
                backoff_limit=backoff_limit,
                ttl_seconds_after_finished=settings.ttl_seconds_after_finished,
                active_deadline_seconds=settings.active_deadline_seconds,
                pod_failure_policy=pod_failure_policy,
                owner_references=owner_references,
                labels=step_labels,
                annotations=step_annotations,
            )

            if (
                startup_interval
                := orchestrator.config.parallel_step_startup_waiting_period
            ):
                nonlocal last_startup_time

                with startup_lock:
                    now = time.time()
                    time_since_last_startup = now - last_startup_time
                    sleep_time = startup_interval - time_since_last_startup
                    if sleep_time > 0:
                        logger.debug(
                            f"Sleeping for {sleep_time} seconds before "
                            f"starting job for step {step_name}."
                        )
                        time.sleep(sleep_time)
                    last_startup_time = now

            kube_utils.create_job(
                batch_api=batch_api,
                namespace=namespace,
                job_manifest=job_manifest,
            )

            Client().create_run_metadata(
                metadata={"step_jobs": {step_name: job_name}},
                resources=[
                    RunMetadataResource(
                        id=pipeline_run.id,
                        type=MetadataResourceTypes.PIPELINE_RUN,
                    )
                ],
            )

            node.metadata["job_name"] = job_name

            return NodeStatus.RUNNING

        def stop_step(node: Node) -> None:
            """Stop a step.

            Args:
                node: The node to stop.
            """
            step_name = node.id

            label_selector = (
                f"run_id={kube_utils.sanitize_label(str(pipeline_run.id))},"
                f"step_name={kube_utils.sanitize_label(node.id)}"
            )

            try:
                jobs = batch_api.list_namespaced_job(
                    namespace=namespace,
                    label_selector=label_selector,
                )
            except Exception as e:
                logger.warning(
                    f"Failed to find jobs for step {step_name}: {e}"
                )
                return

            if not jobs.items:
                logger.debug(f"No jobs found for step {step_name}")
                return

            for job in jobs.items:
                # Check if this is a job related to the correct step
                step_name_annotation = job.metadata.annotations.get(
                    STEP_NAME_ANNOTATION_KEY, None
                )
                if (
                    step_name_annotation is None
                    or step_name_annotation != step_name
                ):
                    continue

                # Check if job is already completed or failed - skip deletion
                job_finished = False
                if job.status.conditions:
                    for condition in job.status.conditions:
                        if (
                            condition.type in ["Complete", "Failed"]
                            and condition.status == "True"
                        ):
                            job_finished = True
                            break

                if job_finished:
                    logger.debug(
                        f"Job {job.metadata.name} for step {step_name} already finished, skipping deletion"
                    )
                    continue

                # Delete the running/pending job
                try:
                    batch_api.delete_namespaced_job(
                        name=job.metadata.name,
                        namespace=namespace,
                        propagation_policy="Foreground",
                    )
                    logger.info(
                        f"Successfully stopped step job: {job.metadata.name}"
                    )
                    break
                except Exception as e:
                    logger.warning(
                        f"Failed to stop step job {job.metadata.name}: {e}"
                    )
                    break

        def check_job_status(node: Node) -> NodeStatus:
            """Check the status of a job.

            Args:
                node: The node to check.

            Returns:
                The status of the node.
            """
            step_name = node.id
            job_name = node.metadata.get("job_name", None)
            if not job_name:
                logger.error(
                    "Missing job name to monitor step `%s`.", step_name
                )
                return NodeStatus.FAILED

            step_config = snapshot.step_configurations[step_name].config
            settings = step_config.settings.get(
                "orchestrator.kubernetes", None
            )
            settings = KubernetesOrchestratorSettings.model_validate(
                settings.model_dump() if settings else {}
            )
            status, error_message = kube_utils.check_job_status(
                batch_api=batch_api,
                core_api=core_api,
                namespace=namespace,
                job_name=job_name,
                fail_on_container_waiting_reasons=settings.fail_on_container_waiting_reasons,
            )
            if status == kube_utils.JobStatus.SUCCEEDED:
                return NodeStatus.COMPLETED
            elif status == kube_utils.JobStatus.FAILED:
                logger.error(
                    "Job for step `%s` failed: %s",
                    step_name,
                    error_message,
                )
                return NodeStatus.FAILED
            else:
                return NodeStatus.RUNNING

        def should_interrupt_execution() -> Optional[InterruptMode]:
            """Check if the DAG execution should be interrupted.

            Returns:
                If the DAG execution should be interrupted.
            """
            try:
                run = client.get_pipeline_run(
                    name_id_or_prefix=pipeline_run.id,
                    project=pipeline_run.project_id,
                    hydrate=False,  # We only need status, not full hydration
                )

                if run.status in [
                    ExecutionStatus.STOPPING,
                    ExecutionStatus.STOPPED,
                ]:
                    logger.info(
                        "Stopping DAG execution because pipeline run is in "
                        "`%s` state.",
                        run.status,
                    )
                    return InterruptMode.GRACEFUL
                elif run.status == ExecutionStatus.FAILED:
                    if (
                        snapshot.pipeline_configuration.execution_mode
                        == ExecutionMode.STOP_ON_FAILURE
                    ):
                        logger.info(
                            "Stopping DAG execution because pipeline run is in "
                            "`%s` state.",
                            run.status,
                        )
                        return InterruptMode.GRACEFUL
                    elif (
                        snapshot.pipeline_configuration.execution_mode
                        == ExecutionMode.FAIL_FAST
                    ):
                        logger.info(
                            "Stopping DAG execution because pipeline run is in "
                            "`%s` state.",
                            run.status,
                        )
                        return InterruptMode.FORCE
            except Exception as e:
                logger.warning(
                    "Failed to check pipeline cancellation status: %s", e
                )

            return None

        try:
            nodes_statuses = DagRunner(
                nodes=nodes,
                node_startup_function=start_step_job,
                node_monitoring_function=check_job_status,
                interrupt_function=should_interrupt_execution,
                monitoring_interval=pipeline_settings.job_monitoring_interval,
                monitoring_delay=pipeline_settings.job_monitoring_delay,
                interrupt_check_interval=pipeline_settings.interrupt_check_interval,
                max_parallelism=pipeline_settings.max_parallelism,
                node_stop_function=stop_step,
            ).run()
        finally:
            if (
                orchestrator.config.pass_zenml_token_as_secret
                and snapshot.schedule is None
            ):
                secret_name = orchestrator.get_token_secret_name(snapshot.id)
                try:
                    kube_utils.delete_secret(
                        core_api=core_api,
                        namespace=namespace,
                        secret_name=secret_name,
                    )
                except ApiException as e:
                    logger.error(
                        f"Error cleaning up secret {secret_name}: {e}"
                    )

        try:
            pipeline_failed = False
            failed_step_names = [
                step_name
                for step_name, node_state in nodes_statuses.items()
                if node_state == NodeStatus.FAILED
            ]
            skipped_step_names = [
                step_name
                for step_name, node_state in nodes_statuses.items()
                if node_state == NodeStatus.SKIPPED
            ]

            if failed_step_names:
                logger.error(
                    "The following steps failed: %s",
                    ", ".join(failed_step_names),
                )
            if skipped_step_names:
                logger.error(
                    "The following steps were skipped because some of their "
                    "upstream steps failed: %s",
                    ", ".join(skipped_step_names),
                )

            step_runs = fetch_step_runs_by_names(
                step_run_names=failed_step_names, pipeline_run=pipeline_run
            )

            for step_name, node_state in nodes_statuses.items():
                if node_state != NodeStatus.FAILED:
                    continue

                pipeline_failed = True

                if step_run := step_runs.get(step_name, None):
                    # Try to update the step run status, if it exists and is in
                    # a transient state.
                    if step_run and step_run.status in {
                        ExecutionStatus.INITIALIZING,
                        ExecutionStatus.RUNNING,
                    }:
                        publish_utils.publish_failed_step_run(step_run.id)

            # If any steps failed and the pipeline run is still in a transient
            # state, we need to mark it as failed.
            if pipeline_failed and pipeline_run.status in {
                ExecutionStatus.INITIALIZING,
                ExecutionStatus.RUNNING,
            }:
                publish_utils.publish_failed_pipeline_run(pipeline_run.id)
        except AuthorizationException:
            # If a step of the pipeline failed or all of them completed
            # successfully, the pipeline run will be finished and the API token
            # will be invalidated. We catch this exception and do nothing here,
            # as the pipeline run status will already have been published.
            pass

        logger.info("Orchestrator pod finished.")


if __name__ == "__main__":
    main()



#  Copyright (c) ZenML GmbH 2024. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Runner entrypoint configuration."""

from typing import Any, Dict, List
from uuid import UUID

from zenml.client import Client
from zenml.entrypoints.base_entrypoint_configuration import (
    BaseEntrypointConfiguration,
)
from zenml.execution.pipeline.utils import submit_pipeline

PLACEHOLDER_RUN_ID_OPTION = "placeholder_run_id"


class RunnerEntrypointConfiguration(BaseEntrypointConfiguration):
    """Runner entrypoint configuration."""

    @classmethod
    def get_entrypoint_options(cls) -> Dict[str, bool]:
        """Gets all options required for running with this configuration.

        Returns:
            The superclass options as well as an option for the name of the
            step to run.
        """
        return super().get_entrypoint_options() | {
            PLACEHOLDER_RUN_ID_OPTION: True
        }

    @classmethod
    def get_entrypoint_arguments(
        cls,
        **kwargs: Any,
    ) -> List[str]:
        """Gets all arguments that the entrypoint command should be called with.

        Args:
            **kwargs: Kwargs, must include the placeholder run id.

        Returns:
            The superclass arguments as well as arguments for the placeholder
            run id.
        """
        return super().get_entrypoint_arguments(**kwargs) + [
            f"--{PLACEHOLDER_RUN_ID_OPTION}",
            str(kwargs[PLACEHOLDER_RUN_ID_OPTION]),
        ]

    def run(self) -> None:
        """Run the entrypoint configuration.

        This method runs the pipeline defined by the snapshot given as input
        to the entrypoint configuration.
        """
        snapshot = self.snapshot
        placeholder_run_id = UUID(
            self.entrypoint_args[PLACEHOLDER_RUN_ID_OPTION]
        )
        placeholder_run = Client().get_pipeline_run(placeholder_run_id)

        stack = Client().active_stack
        assert snapshot.stack and stack.id == snapshot.stack.id

        submit_pipeline(
            snapshot=snapshot,
            stack=stack,
            placeholder_run=placeholder_run,
        )



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Dynamic pipeline execution utilities."""

from typing import (
    Generic,
    TypeVar,
)

T = TypeVar("T")


class _Unmapped(Generic[T]):
    """Wrapper class for inputs that should not be mapped over."""

    def __init__(self, value: T):
        """Initialize the wrapper.

        Args:
            value: The value to wrap.
        """
        self.value = value


def unmapped(value: T) -> _Unmapped[T]:
    """Helper function to pass an input without mapping over it.

    Wrap any step input with this function and then pass it to `step.map(...)`
    to pass the full value to all steps.

    Args:
        value: The value to wrap.

    Returns:
        The wrapped value.
    """
    return _Unmapped(value)



"""Release [0.56.1].

Revision ID: 0.56.1
Revises: 0.56.0
Create Date: 2024-03-21 14:50:20.869911

"""

# revision identifiers, used by Alembic.
revision = "0.56.1"
down_revision = "0.56.0"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass



#  Copyright (c) ZenML GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""ZenML specific exception definitions."""

from typing import Dict, Optional


class ZenMLBaseException(Exception):
    """Base exception for all ZenML Exceptions."""

    def __init__(
        self,
        message: Optional[str] = None,
        url: Optional[str] = None,
    ):
        """The BaseException used to format messages displayed to the user.

        Args:
            message: Message with details of exception. This message
                     will be appended with another message directing user to
                     `url` for more information. If `None`, then default
                     Exception behavior is used.
            url: URL to point to in exception message. If `None`, then no url
                 is appended.
        """
        if message and url:
            message += f" For more information, visit {url}."
        super().__init__(message)


class InitializationException(ZenMLBaseException):
    """Raised when an error occurred during initialization of a ZenML repository."""


class AuthorizationException(ZenMLBaseException):
    """Raised when an authorization error occurred while trying to access a ZenML resource ."""


class CredentialsNotValid(AuthorizationException):
    """Raised when the credentials provided are invalid.

    This is a subclass of AuthorizationException and should only be raised when
    the authentication credentials are invalid (e.g. expired API token, invalid
    username/password, invalid signature). If caught by the ZenML client, it
    will trigger an invalidation of the currently cached API token and a
    re-authentication flow.
    """


class DoesNotExistException(ZenMLBaseException):
    """Raises exception when the entity does not exist in the system but an action is being done that requires it to be present."""

    def __init__(self, message: str):
        """Initializes the exception.

        Args:
            message: Message with details of exception.
        """
        super().__init__(message)


class StepInterfaceError(ZenMLBaseException):
    """Raises exception when interacting with the Step interface in an unsupported way."""


class MaterializerInterfaceError(ZenMLBaseException):
    """Raises exception when interacting with the Materializer interface in an unsupported way."""


class StepContextError(ZenMLBaseException):
    """Raises exception when interacting with a StepContext in an unsupported way."""


class StackComponentInterfaceError(ZenMLBaseException):
    """Raises exception when interacting with the stack components in an unsupported way."""


class ArtifactStoreInterfaceError(ZenMLBaseException):
    """Raises exception when interacting with the Artifact Store interface in an unsupported way."""


class IntegrationError(ZenMLBaseException):
    """Raises exceptions when a requested integration can not be activated."""


class ValidationError(ZenMLBaseException):
    """Raised when the Model passed to the ZenStore."""


class EntityExistsError(ZenMLBaseException):
    """Raised when trying to register an entity that already exists."""


class EntityCreationError(ZenMLBaseException, RuntimeError):
    """Raised when failing to create an entity."""


class WebhookInactiveError(ZenMLBaseException):
    """Raised when source is inactive."""


class StackValidationError(ZenMLBaseException):
    """Raised when a stack configuration is not valid."""


class GitNotFoundError(ImportError):
    """Raised when ZenML CLI is used to interact with examples on a machine with no git installation."""


class IllegalOperationError(ZenMLBaseException):
    """Raised when an illegal operation is attempted."""


class RunStoppedException(ZenMLBaseException):
    """Raised when a ZenML pipeline run gets stopped by the user."""


class RunInterruptedException(ZenMLBaseException):
    """Raised when a ZenML step gets interrupted for an unknown reason."""


class MethodNotAllowedError(ZenMLBaseException):
    """Raised when the server does not allow a request method."""


class SettingsResolvingError(ZenMLBaseException):
    """Raised when resolving settings failed."""


class InputResolutionError(ZenMLBaseException):
    """Raised when step input resolving failed."""


class SubscriptionUpgradeRequiredError(ZenMLBaseException):
    """Raised when user tries to perform an action outside their current subscription tier."""


class HydrationError(ZenMLBaseException):
    """Raised when the model hydration failed."""


class ZenKeyError(KeyError):
    """Specialized key error which allows error messages with line breaks."""

    def __init__(self, message: str) -> None:
        """Initialization.

        Args:
            message:str, the error message
        """
        self.message = message

    def __str__(self) -> str:
        """String function.

        Returns:
            the error message
        """
        return self.message


class OAuthError(ValueError):
    """OAuth2 error."""

    def __init__(
        self,
        error: str,
        status_code: int = 400,
        error_description: Optional[str] = None,
        error_uri: Optional[str] = None,
    ) -> None:
        """Initializes the OAuthError.

        Args:
            status_code: HTTP status code.
            error: Error code.
            error_description: Error description.
            error_uri: Error URI.
        """
        self.status_code = status_code
        self.error = error
        self.error_description = error_description
        self.error_uri = error_uri

    def to_dict(self) -> Dict[str, Optional[str]]:
        """Returns the OAuthError as a dictionary.

        Returns:
            The OAuthError as a dictionary.
        """
        return {
            "error": self.error,
            "error_description": self.error_description,
            "error_uri": self.error_uri,
        }

    def __str__(self) -> str:
        """String function.

        Returns:
            the error message
        """
        return f"{self.error}: {self.error_description or ''}"


class SecretsStoreNotConfiguredError(NotImplementedError):
    """Raised when a secrets store is not configured."""


class BackupSecretsStoreNotConfiguredError(NotImplementedError):
    """Raised when a backup secrets store is not configured."""


class CustomFlavorImportError(ImportError):
    """Raised when failing to import a custom flavor."""


class MaxConcurrentTasksError(ZenMLBaseException):
    """Raised when the maximum number of concurrent tasks is reached."""


class RunMonitoringError(ZenMLBaseException):
    """Raised when an error occurs while monitoring a pipeline run."""

    def __init__(
        self,
        original_exception: BaseException,
    ) -> None:
        """Initializes the error.

        Args:
            original_exception: The original exception that occurred while
                monitoring the pipeline run.
        """
        self.original_exception = original_exception


class HookValidationException(ZenMLBaseException):
    """Exception raised when hook validation fails."""


class HookExecutionException(ZenMLBaseException):
    """Exception raised when hook execution fails."""



#  Copyright (c) ZenML GmbH 2023. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""ZenML logging handler."""

import asyncio
import logging
import os
import queue
import re
import threading
import time
from contextlib import contextmanager, nullcontext
from contextvars import ContextVar
from datetime import datetime
from types import TracebackType
from typing import (
    Any,
    Generator,
    Iterator,
    List,
    Optional,
    Type,
    Union,
)
from uuid import UUID, uuid4

from pydantic import BaseModel, Field

from zenml.artifact_stores import BaseArtifactStore
from zenml.artifacts.utils import _load_artifact_store
from zenml.client import Client
from zenml.constants import (
    ENV_ZENML_DISABLE_PIPELINE_LOGS_STORAGE,
    ENV_ZENML_DISABLE_STEP_NAMES_IN_LOGS,
    LOGS_MERGE_INTERVAL_SECONDS,
    LOGS_STORAGE_MAX_QUEUE_SIZE,
    LOGS_STORAGE_QUEUE_TIMEOUT,
    LOGS_WRITE_INTERVAL_SECONDS,
    handle_bool_env_var,
)
from zenml.enums import LoggingLevels
from zenml.exceptions import DoesNotExistException
from zenml.logger import (
    get_logger,
    get_storage_log_level,
    logging_handlers,
    step_names_in_console,
)
from zenml.models import (
    LogsRequest,
    LogsResponse,
    PipelineRunUpdate,
    PipelineSnapshotResponse,
)
from zenml.utils.io_utils import sanitize_remote_path
from zenml.utils.time_utils import utc_now
from zenml.zen_stores.base_zen_store import BaseZenStore

logger = get_logger(__name__)

# Context variables
redirected: ContextVar[bool] = ContextVar("redirected", default=False)

ansi_escape = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")

LOGS_EXTENSION = ".log"
PIPELINE_RUN_LOGS_FOLDER = "pipeline_runs"

# Maximum number of log entries to return in a single request
MAX_ENTRIES_PER_REQUEST = 20000
# Maximum size of a single log message in bytes (5KB)
DEFAULT_MESSAGE_SIZE = 5 * 1024


class LogEntry(BaseModel):
    """A structured log entry with parsed information."""

    message: str = Field(description="The log message content")
    name: Optional[str] = Field(
        default=None,
        description="The name of the logger",
    )
    level: Optional[LoggingLevels] = Field(
        default=None,
        description="The log level",
    )
    timestamp: Optional[datetime] = Field(
        default=None,
        description="When the log was created",
    )
    module: Optional[str] = Field(
        default=None, description="The module that generated this log entry"
    )
    filename: Optional[str] = Field(
        default=None,
        description="The name of the file that generated this log entry",
    )
    lineno: Optional[int] = Field(
        default=None, description="The fileno that generated this log entry"
    )
    chunk_index: int = Field(
        default=0,
        description="The index of the chunk in the log entry",
    )
    total_chunks: int = Field(
        default=1,
        description="The total number of chunks in the log entry",
    )
    id: UUID = Field(
        default_factory=uuid4,
        description="The unique identifier of the log entry",
    )


class ArtifactStoreHandler(logging.Handler):
    """Handler that writes log messages to artifact store storage."""

    def __init__(self, storage: "PipelineLogsStorage"):
        """Initialize the handler with a storage instance.

        Args:
            storage: The PipelineLogsStorage instance to write to.
        """
        super().__init__()
        self.storage = storage

        # Get storage log level from environment
        self.setLevel(get_storage_log_level().value)

    def emit(self, record: logging.LogRecord) -> None:
        """Emit a log record to the storage.

        Args:
            record: The log record to emit.
        """
        try:
            # Get level enum
            level = LoggingLevels.__members__.get(record.levelname.upper())

            # Get the message
            message = self.format(record)
            message = remove_ansi_escape_codes(message).rstrip()

            # Check if message needs to be chunked
            message_bytes = message.encode("utf-8")
            if len(message_bytes) <= DEFAULT_MESSAGE_SIZE:
                # Message is small enough, emit as-is
                log_record = LogEntry.model_construct(
                    message=message,
                    name=record.name,
                    level=level,
                    timestamp=utc_now(tz_aware=True),
                    module=record.module,
                    filename=record.filename,
                    lineno=record.lineno,
                )
                json_line = log_record.model_dump_json(exclude_none=True)
                self.storage.write(json_line)
            else:
                # Message is too large, split into chunks and emit each one
                chunks = self._split_to_chunks(message)
                entry_id = uuid4()
                for i, chunk in enumerate(chunks):
                    log_record = LogEntry.model_construct(
                        message=chunk,
                        name=record.name,
                        level=level,
                        module=record.module,
                        filename=record.filename,
                        lineno=record.lineno,
                        timestamp=utc_now(tz_aware=True),
                        chunk_index=i,
                        total_chunks=len(chunks),
                        id=entry_id,
                    )

                    json_line = log_record.model_dump_json(exclude_none=True)
                    self.storage.write(json_line)
        except Exception:
            pass

    def _split_to_chunks(self, message: str) -> List[str]:
        """Split a large message into chunks.

        Args:
            message: The message to split.

        Returns:
            A list of message chunks.
        """
        # Calculate how many chunks we need
        message_bytes = message.encode("utf-8")

        # Split the message into chunks, handling UTF-8 boundaries
        chunks = []
        start = 0

        while start < len(message_bytes):
            # Calculate the end position for this chunk
            end = min(start + DEFAULT_MESSAGE_SIZE, len(message_bytes))

            # Try to decode the chunk, backing up if we hit a UTF-8 boundary issue
            while end > start:
                chunk_bytes = message_bytes[start:end]
                try:
                    chunk_text = chunk_bytes.decode("utf-8")
                    chunks.append(chunk_text)
                    break
                except UnicodeDecodeError:
                    # If we can't decode, try a smaller chunk
                    end -= 1
            else:
                # If we can't decode anything, use replacement characters
                end = min(start + DEFAULT_MESSAGE_SIZE, len(message_bytes))
                chunks.append(
                    message_bytes[start:end].decode("utf-8", errors="replace")
                )

            start = end

        return chunks


def remove_ansi_escape_codes(text: str) -> str:
    """Auxiliary function to remove ANSI escape codes from a given string.

    Args:
        text: the input string

    Returns:
        the version of the input string where the escape codes are removed.
    """
    return ansi_escape.sub("", text)


def parse_log_entry(log_line: str) -> Optional[LogEntry]:
    """Parse a single log entry into a LogEntry object.

    Handles two formats:
    1. JSON format: {"timestamp": "...", "level": "...", "message": "...", "location": "..."}
       Uses Pydantic's model_validate_json for automatic parsing and validation.
    2. Plain text: Any other text (defaults to INFO level)

    Args:
        log_line: A single log line to parse

    Returns:
        LogEntry object. For JSON logs, all fields are validated and parsed automatically.
        For plain text logs, only message is populated with INFO level default.
        Returns None only for empty lines.
    """
    line = log_line.strip()
    if not line:
        return None

    if line.startswith("{") and line.endswith("}"):
        try:
            return LogEntry.model_validate_json(line)
        except Exception:
            pass

    old_format = re.search(
        r"^\[(\d{4}-\d{2}-\d{2}\s+\d{2}:\d{2}:\d{2})\s+UTC\]", line
    )

    timestamp = None
    if old_format:
        timestamp = old_format.group(1) + "Z"
        line = line.replace(old_format.group(0), "").strip()

    return LogEntry(
        message=line,
        name=None,
        level=LoggingLevels.INFO,
        timestamp=timestamp,
    )


def prepare_logs_uri(
    artifact_store: "BaseArtifactStore",
    step_name: Optional[str] = None,
    log_key: Optional[str] = None,
) -> str:
    """Generates and prepares a URI for the log file or folder for a step.

    Args:
        artifact_store: The artifact store on which the artifact will be stored.
        step_name: Name of the step. Skipped for global pipeline run logs.
        log_key: The unique identification key of the log file.

    Returns:
        The URI of the log storage (file or folder).
    """
    if log_key is None:
        log_key = str(uuid4())

    subfolder = step_name or PIPELINE_RUN_LOGS_FOLDER
    logs_base_uri = os.path.join(artifact_store.path, subfolder, "logs")

    # Create the dir
    if not artifact_store.exists(logs_base_uri):
        artifact_store.makedirs(logs_base_uri)

    # Delete the file if it already exists
    if artifact_store.config.IS_IMMUTABLE_FILESYSTEM:
        logs_uri = os.path.join(logs_base_uri, log_key)
        if artifact_store.exists(logs_uri):
            logger.warning(
                f"Logs directory {logs_uri} already exists! Removing old log directory..."
            )
            artifact_store.rmtree(logs_uri)

        artifact_store.makedirs(logs_uri)
    else:
        logs_uri = os.path.join(logs_base_uri, f"{log_key}{LOGS_EXTENSION}")
        if artifact_store.exists(logs_uri):
            logger.warning(
                f"Logs file {logs_uri} already exists! Removing old log file..."
            )
            artifact_store.remove(logs_uri)

    return sanitize_remote_path(logs_uri)


def fetch_log_records(
    zen_store: "BaseZenStore",
    artifact_store_id: Union[str, UUID],
    logs_uri: str,
) -> List[LogEntry]:
    """Fetches log entries.

    Args:
        zen_store: The store in which the artifact is stored.
        artifact_store_id: The ID of the artifact store.
        logs_uri: The URI of the artifact (file or directory).

    Returns:
        List of log entries.
    """
    log_entries = []

    for line in _stream_logs_line_by_line(
        zen_store, artifact_store_id, logs_uri
    ):
        if log_entry := parse_log_entry(line):
            log_entries.append(log_entry)

        if len(log_entries) >= MAX_ENTRIES_PER_REQUEST:
            break

    return log_entries


def _stream_logs_line_by_line(
    zen_store: "BaseZenStore",
    artifact_store_id: Union[str, UUID],
    logs_uri: str,
) -> Iterator[str]:
    """Stream logs line by line without loading the entire file into memory.

    This generator yields log lines one by one, handling both single files
    and directories with multiple log files.

    Args:
        zen_store: The store in which the artifact is stored.
        artifact_store_id: The ID of the artifact store.
        logs_uri: The URI of the log file or directory.

    Yields:
        Individual log lines as strings.

    Raises:
        DoesNotExistException: If the artifact does not exist in the artifact store.
    """
    artifact_store = _load_artifact_store(artifact_store_id, zen_store)

    try:
        if not artifact_store.isdir(logs_uri):
            # Single file case
            with artifact_store.open(logs_uri, "r") as file:
                for line in file:
                    yield line.rstrip("\n\r")
        else:
            # Directory case - may contain multiple log files
            files = artifact_store.listdir(logs_uri)
            if not files:
                raise DoesNotExistException(
                    f"Folder '{logs_uri}' is empty in artifact store "
                    f"'{artifact_store.name}'."
                )

            # Sort files to read them in order
            files.sort()

            for file in files:
                file_path = os.path.join(logs_uri, str(file))
                with artifact_store.open(file_path, "r") as f:
                    for line in f:
                        yield line.rstrip("\n\r")
    finally:
        artifact_store.cleanup()


class PipelineLogsStorage:
    """Helper class which buffers and stores logs to a given URI using a background thread."""

    def __init__(
        self,
        logs_uri: str,
        artifact_store: "BaseArtifactStore",
        max_queue_size: int = LOGS_STORAGE_MAX_QUEUE_SIZE,
        queue_timeout: int = LOGS_STORAGE_QUEUE_TIMEOUT,
        write_interval: int = LOGS_WRITE_INTERVAL_SECONDS,
        merge_files_interval: int = LOGS_MERGE_INTERVAL_SECONDS,
    ) -> None:
        """Initialization.

        Args:
            logs_uri: the URI of the log file or folder.
            artifact_store: Artifact Store from the current step context
            max_queue_size: maximum number of individual messages to queue.
            queue_timeout: timeout in seconds for putting items in queue when full.
                - Positive value: Wait N seconds, then drop logs if queue still full
                - Negative value: Block indefinitely until queue has space (never drop logs)
            write_interval: the amount of seconds before the created files
                get written to the artifact store.
            merge_files_interval: the amount of seconds before the created files
                get merged into a single file.
        """
        # Parameters
        self.logs_uri = logs_uri
        self.max_queue_size = max_queue_size
        self.queue_timeout = queue_timeout
        self.write_interval = write_interval
        self.merge_files_interval = merge_files_interval

        # State
        self.artifact_store = artifact_store

        # Immutable filesystems state
        self.last_merge_time = time.time()

        # Queue and log storage thread for async processing
        self.log_queue: queue.Queue[str] = queue.Queue(maxsize=max_queue_size)
        self.log_storage_thread: Optional[threading.Thread] = None
        self.shutdown_event = threading.Event()
        self.merge_event = threading.Event()

        # Start the log storage thread
        self._start_log_storage_thread()

    def _start_log_storage_thread(self) -> None:
        """Start the log storage thread for processing log queue."""
        if (
            self.log_storage_thread is None
            or not self.log_storage_thread.is_alive()
        ):
            self.log_storage_thread = threading.Thread(
                target=self._log_storage_worker,
                name="LogsStorage-Worker",
            )
            self.log_storage_thread.start()

    def _process_log_queue(self, force_merge: bool = False) -> None:
        """Write and merge logs to the artifact store using time-based batching.

        Args:
            force_merge: Whether to force merge the logs.
        """
        try:
            messages = []

            # Get first message (blocking with timeout)
            try:
                first_message = self.log_queue.get(timeout=1)
                messages.append(first_message)
            except queue.Empty:
                return

            # Get any remaining messages without waiting (drain quickly)
            while True:
                try:
                    additional_message = self.log_queue.get_nowait()
                    messages.append(additional_message)
                except queue.Empty:
                    break

            # Write the messages to the artifact store
            if messages:
                self.write_buffer(messages)

            # Merge the log files if needed
            if (
                self._is_merge_needed
                or self.merge_event.is_set()
                or force_merge
            ):
                self.merge_event.clear()

                self.merge_log_files(merge_all_files=force_merge)

        except Exception as e:
            logger.error("Error in log storage thread: %s", e)
        finally:
            for _ in messages:
                self.log_queue.task_done()

            # Wait for the next write interval or until shutdown is requested
            self.shutdown_event.wait(timeout=self.write_interval)

    def _log_storage_worker(self) -> None:
        """Log storage thread worker that processes the log queue."""
        # Process the log queue until shutdown is requested
        while not self.shutdown_event.is_set():
            self._process_log_queue()

        # Shutdown requested - drain remaining queue items and merge log files
        self._process_log_queue(force_merge=True)

    def _shutdown_log_storage_thread(self, timeout: int = 5) -> None:
        """Shutdown the log storage thread gracefully.

        Args:
            timeout: Maximum time to wait for thread shutdown.
        """
        if self.log_storage_thread and self.log_storage_thread.is_alive():
            # Then signal the worker to begin graceful shutdown
            self.shutdown_event.set()

            # Wait for thread to finish (it will drain the queue automatically)
            self.log_storage_thread.join(timeout=timeout)

    def write(self, text: str) -> None:
        """Main write method that sends individual messages directly to queue.

        Args:
            text: the incoming string.
        """
        # Skip empty lines
        if text == "\n":
            return

        # If the current thread is the log storage thread, do nothing
        # to prevent recursion when the storage thread itself generates logs
        if (
            self.log_storage_thread
            and threading.current_thread() == self.log_storage_thread
        ):
            return

        # If the current thread is the fsspec IO thread, do nothing
        if self._is_fsspec_io_thread:
            return

        try:
            # Send individual message directly to queue
            if not self.shutdown_event.is_set():
                try:
                    if self.queue_timeout < 0:
                        # Negative timeout = block indefinitely until queue has space
                        # Guarantees no log loss but may hang application
                        self.log_queue.put(text)
                    else:
                        # Positive timeout = wait specified time then drop logs
                        # Prevents application hanging but may lose logs
                        self.log_queue.put(text, timeout=self.queue_timeout)
                except queue.Full:
                    # This only happens with positive timeout
                    # Queue is full - just skip this message to avoid blocking
                    # Better to drop logs than hang the application
                    pass

        except Exception:
            # Silently ignore errors to prevent recursion
            pass

    @property
    def _is_merge_needed(self) -> bool:
        """Checks whether the log files need to be merged.

        Returns:
            whether the log files need to be merged.
        """
        return (
            self.artifact_store.config.IS_IMMUTABLE_FILESYSTEM
            and time.time() - self.last_merge_time > self.merge_files_interval
        )

    @property
    def _is_fsspec_io_thread(self) -> bool:
        """Checks if the current thread is the fsspec IO thread.

        Returns:
            whether the current thread is the fsspec IO thread.
        """
        # Most artifact stores are based on fsspec, which converts between
        # sync and async operations by using a separate AIO thread.
        # It may happen that the fsspec call itself will log something,
        # which will trigger this method, which may then use fsspec again,
        # causing a "Calling sync() from within a running loop" error, because
        # the fsspec library does not expect sync calls being made as a result
        # of a logging call made by itself.
        # To avoid this, we simply check if we're running in the fsspec AIO
        # thread and skip the save if that's the case.
        try:
            return (
                asyncio.events.get_running_loop() is not None
                and threading.current_thread().name == "fsspecIO"
            )
        except RuntimeError:
            # No running loop
            return False

    def _get_timestamped_filename(self, suffix: str = "") -> str:
        """Returns a timestamped filename.

        Args:
            suffix: optional suffix for the file name

        Returns:
            The timestamped filename.
        """
        return f"{time.time()}{suffix}{LOGS_EXTENSION}"

    def write_buffer(self, buffer_to_write: List[str]) -> None:
        """Write the given buffer to file. This runs in the log storage thread.

        Args:
            buffer_to_write: The buffer contents to write to file.
        """
        if not buffer_to_write:
            return

        try:
            # If the artifact store is immutable, write the buffer to a new file
            if self.artifact_store.config.IS_IMMUTABLE_FILESYSTEM:
                _logs_uri = self._get_timestamped_filename()
                with self.artifact_store.open(
                    os.path.join(
                        self.logs_uri,
                        _logs_uri,
                    ),
                    "w",
                ) as file:
                    for message in buffer_to_write:
                        file.write(f"{message}\n")

            # If the artifact store is mutable, append the buffer to the existing file
            else:
                with self.artifact_store.open(self.logs_uri, "a") as file:
                    for message in buffer_to_write:
                        file.write(f"{message}\n")
                self.artifact_store._remove_previous_file_versions(
                    self.logs_uri
                )

        except Exception as e:
            logger.error("Error in log storage thread: %s", e)

    def merge_log_files(self, merge_all_files: bool = False) -> None:
        """Merges all log files into one in the given URI.

        Called on the logging context exit.

        Args:
            merge_all_files: whether to merge all files or only raw files
        """
        from zenml.artifacts.utils import (
            _load_file_from_artifact_store,
        )

        # If the artifact store is immutable, merge the log files
        if self.artifact_store.config.IS_IMMUTABLE_FILESYSTEM:
            merged_file_suffix = "_merged"
            files_ = self.artifact_store.listdir(self.logs_uri)
            if not merge_all_files:
                # already merged files will not be merged again
                files_ = [
                    f for f in files_ if merged_file_suffix not in str(f)
                ]
            file_name_ = self._get_timestamped_filename(
                suffix=merged_file_suffix
            )
            if len(files_) > 1:
                files_.sort()
                logger.debug("Log files count: %s", len(files_))

                missing_files = set()
                # dump all logs to a local file first
                with self.artifact_store.open(
                    os.path.join(self.logs_uri, file_name_), "w"
                ) as merged_file:
                    for file in files_:
                        try:
                            merged_file.write(
                                str(
                                    _load_file_from_artifact_store(
                                        os.path.join(self.logs_uri, str(file)),
                                        artifact_store=self.artifact_store,
                                        mode="r",
                                    )
                                )
                            )
                        except DoesNotExistException:
                            missing_files.add(file)

                # clean up left over files
                for file in files_:
                    if file not in missing_files:
                        self.artifact_store.remove(
                            os.path.join(self.logs_uri, str(file))
                        )

            # Update the last merge time
            self.last_merge_time = time.time()

    def send_merge_event(self) -> None:
        """Send a merge event to the log storage thread."""
        self.merge_event.set()


class PipelineLogsStorageContext:
    """Context manager which collects logs during pipeline run execution."""

    def __init__(
        self,
        logs_uri: str,
        artifact_store: "BaseArtifactStore",
        prepend_step_name: bool = True,
    ) -> None:
        """Initializes and prepares a storage object.

        Args:
            logs_uri: the URI of the logs file.
            artifact_store: Artifact Store from the current pipeline run context.
            prepend_step_name: Whether to prepend the step name to the logs.
        """
        # Create the storage object
        self.storage = PipelineLogsStorage(
            logs_uri=logs_uri, artifact_store=artifact_store
        )

        # Create the handler object
        self.artifact_store_handler: ArtifactStoreHandler = (
            ArtifactStoreHandler(self.storage)
        )

        # Additional configuration
        self.prepend_step_name = prepend_step_name
        self.original_step_names_in_console: Optional[bool] = None
        self._original_root_level: Optional[int] = None

    def __enter__(self) -> "PipelineLogsStorageContext":
        """Enter condition of the context manager.

        Registers an ArtifactStoreHandler for log storage.

        Returns:
            self
        """
        # Add handler to root logger
        root_logger = logging.getLogger()
        root_logger.addHandler(self.artifact_store_handler)

        # Set root logger level to minimum of all active handlers
        # This ensures records can reach any handler that needs them
        self._original_root_level = root_logger.level
        handler_levels = [handler.level for handler in root_logger.handlers]

        # Set root logger to the minimum level among all handlers
        min_level = min(handler_levels)
        if min_level < root_logger.level:
            root_logger.setLevel(min_level)

        # Add handler to context variables for print() capture
        logging_handlers.add(self.artifact_store_handler)

        # Save the current step names context variable state
        self.original_step_names_in_console = step_names_in_console.get()

        # Set the step names context variable
        step_names_disabled = handle_bool_env_var(
            ENV_ZENML_DISABLE_STEP_NAMES_IN_LOGS, default=False
        )

        if step_names_disabled or not self.prepend_step_name:
            # Step names are disabled through the env or they are disabled in the config
            step_names_in_console.set(False)
        else:
            # Otherwise, set it True (default)
            step_names_in_console.set(True)

        redirected.set(True)
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        """Exit condition of the context manager.

        Args:
            exc_type: The class of the exception
            exc_val: The instance of the exception
            exc_tb: The traceback of the exception

        Removes the handler from loggers and context variables.
        """
        if exc_type is not None:
            # Write the exception and its traceback to the logs
            self.artifact_store_handler.emit(
                logging.LogRecord(
                    name="exception",
                    level=logging.ERROR,
                    pathname="",
                    lineno=0,
                    msg="An exception has occurred.",
                    args=(),
                    exc_info=(exc_type, exc_val, exc_tb) if exc_val else None,
                )
            )

        # Remove handler from root logger and restore original level
        root_logger = logging.getLogger()

        # Check if handler is still in the root logger before removing
        if self.artifact_store_handler in root_logger.handlers:
            root_logger.removeHandler(self.artifact_store_handler)

        # Restore original root logger level
        if self._original_root_level is not None:
            root_logger.setLevel(self._original_root_level)

        # Remove handler from context variables
        logging_handlers.remove(self.artifact_store_handler)

        # Shutdown thread (it will automatically drain queue and merge files)
        try:
            self.storage._shutdown_log_storage_thread()
        except Exception:
            pass

        # Restore the original step names context variable state
        if self.original_step_names_in_console is not None:
            step_names_in_console.set(self.original_step_names_in_console)


def setup_orchestrator_logging(
    run_id: UUID,
    snapshot: "PipelineSnapshotResponse",
    logs_response: Optional[LogsResponse] = None,
) -> Any:
    """Set up logging for an orchestrator environment.

    This function can be reused by different orchestrators to set up
    consistent logging behavior.

    Args:
        run_id: The pipeline run ID.
        snapshot: The snapshot of the pipeline run.
        logs_response: The logs response to continue from.

    Returns:
        The logs context (PipelineLogsStorageContext)
    """
    try:
        logging_enabled = True

        if handle_bool_env_var(ENV_ZENML_DISABLE_PIPELINE_LOGS_STORAGE, False):
            logging_enabled = False
        else:
            if (
                snapshot.pipeline_configuration.enable_pipeline_logs
                is not None
            ):
                logging_enabled = (
                    snapshot.pipeline_configuration.enable_pipeline_logs
                )

        if not logging_enabled:
            return nullcontext()

        # Fetch the active stack
        client = Client()
        active_stack = client.active_stack

        if logs_response:
            logs_uri = logs_response.uri
        else:
            logs_uri = prepare_logs_uri(
                artifact_store=active_stack.artifact_store,
            )
            logs_model = LogsRequest(
                uri=logs_uri,
                source="orchestrator",
                artifact_store_id=active_stack.artifact_store.id,
            )

            # Add orchestrator logs to the pipeline run
            try:
                run_update = PipelineRunUpdate(add_logs=[logs_model])
                client.zen_store.update_run(
                    run_id=run_id, run_update=run_update
                )
            except Exception as e:
                logger.error(
                    f"Failed to add orchestrator logs to the run {run_id}: {e}"
                )
                raise e

        return PipelineLogsStorageContext(
            logs_uri=logs_uri,
            artifact_store=active_stack.artifact_store,
            prepend_step_name=False,
        )
    except Exception as e:
        logger.error(
            f"Failed to setup orchestrator logging for run {run_id}: {e}"
        )
        return nullcontext()


@contextmanager
def setup_pipeline_logging(
    source: str,
    snapshot: "PipelineSnapshotResponse",
    run_id: Optional[UUID] = None,
    logs_response: Optional[LogsResponse] = None,
) -> Generator[Optional[LogsRequest], None, None]:
    """Set up logging for a pipeline run.

    Args:
        source: The log source.
        snapshot: The snapshot of the pipeline run.
        run_id: The ID of the pipeline run.
        logs_response: The logs response to continue from.

    Raises:
        Exception: If updating the run with the logs request fails.

    Yields:
        The logs request.
    """
    logging_enabled = True

    if handle_bool_env_var(ENV_ZENML_DISABLE_PIPELINE_LOGS_STORAGE, False):
        logging_enabled = False
    elif snapshot.pipeline_configuration.enable_pipeline_logs is not None:
        logging_enabled = snapshot.pipeline_configuration.enable_pipeline_logs

    if logging_enabled:
        client = Client()
        artifact_store = client.active_stack.artifact_store
        logs_model = None

        if logs_response:
            logs_uri = logs_response.uri
        else:
            logs_uri = prepare_logs_uri(
                artifact_store=artifact_store,
            )
            logs_model = LogsRequest(
                uri=logs_uri,
                source=source,
                artifact_store_id=artifact_store.id,
            )

            if run_id:
                try:
                    run_update = PipelineRunUpdate(add_logs=[logs_model])
                    client.zen_store.update_run(
                        run_id=run_id, run_update=run_update
                    )
                except Exception as e:
                    logger.error(
                        f"Failed to add logs to the run {run_id}: {e}"
                    )
                    raise e

        logging_context = PipelineLogsStorageContext(
            logs_uri=logs_uri,
            artifact_store=artifact_store,
            prepend_step_name=False,
        )
        with logging_context:
            yield logs_model
    else:
        yield None



#  Copyright (c) ZenML GmbH 2023. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Utility functions for handling environment variables."""

import contextlib
import os
import re
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterator,
    List,
    Match,
    Optional,
    TypeVar,
    cast,
)

from zenml.client import Client
from zenml.logger import get_logger
from zenml.utils import string_utils

if TYPE_CHECKING:
    from zenml.config.step_configurations import StepConfiguration
    from zenml.stack import Stack

logger = get_logger(__name__)

V = TypeVar("V", bound=Any)
ENV_VARIABLE_PLACEHOLDER_PATTERN = re.compile(pattern=r"\$\{([a-zA-Z0-9_]+)\}")

ENV_VAR_CHUNK_SUFFIX = "_CHUNK_"


def split_environment_variables(
    size_limit: int,
    env: Optional[Dict[str, str]] = None,
) -> None:
    """Split long environment variables into chunks.

    Splits the input environment variables with values that exceed the supplied
    maximum length into individual components. The input environment variables
    are modified in-place.

    Args:
        size_limit: Maximum length of an environment variable value.
        env: Input environment variables dictionary. If not supplied, the
            OS environment variables are used.

    Raises:
        RuntimeError: If an environment variable value is too large and requires
            more than 10 chunks.
    """
    if env is None:
        env = cast(Dict[str, str], os.environ)

    for key, value in env.copy().items():
        if len(value) <= size_limit:
            continue

        # We keep the number of chunks to a maximum of 10 to avoid generating
        # too many environment variables chunks and also to make the
        # reconstruction easier to implement
        if len(value) > size_limit * 10:
            raise RuntimeError(
                f"Environment variable {key} exceeds the maximum length of "
                f"{size_limit * 10} characters."
            )

        env.pop(key)

        # Split the environment variable into chunks
        chunks = [
            value[i : i + size_limit] for i in range(0, len(value), size_limit)
        ]
        for i, chunk in enumerate(chunks):
            env[f"{key}{ENV_VAR_CHUNK_SUFFIX}{i}"] = chunk


def reconstruct_environment_variables(
    env: Optional[Dict[str, str]] = None,
) -> None:
    """Reconstruct environment variables that were split into chunks.

    Reconstructs the environment variables with values that were split into
    individual chunks because they were too large. The input environment
    variables are modified in-place.

    Args:
        env: Input environment variables dictionary. If not supplied, the OS
            environment variables are used.
    """
    if env is None:
        env = cast(Dict[str, str], os.environ)

    chunks: Dict[str, List[str]] = {}
    for key in env.keys():
        if not key[:-1].endswith(ENV_VAR_CHUNK_SUFFIX):
            continue

        # Collect all chunks of the same environment variable
        original_key = key[: -(len(ENV_VAR_CHUNK_SUFFIX) + 1)]
        chunks.setdefault(original_key, [])
        chunks[original_key].append(key)

    # Reconstruct the environment variables from their chunks
    for key, chunk_keys in chunks.items():
        chunk_keys.sort()
        value = "".join([env[key] for key in chunk_keys])
        env[key] = value

        # Remove the chunk environment variables
        for key in chunk_keys:
            env.pop(key)


def substitute_env_variable_placeholders(
    value: V, raise_when_missing: bool = True
) -> V:
    """Substitute environment variable placeholders in an object.

    Args:
        value: The object in which to substitute the placeholders.
        raise_when_missing: If True, an exception will be raised when an
            environment variable is missing. Otherwise, a warning will be logged
            instead.

    Returns:
        The object with placeholders substituted.
    """

    def _replace_with_env_variable_value(match: Match[str]) -> str:
        key = match.group(1)
        if key in os.environ:
            return os.environ[key]
        else:
            if raise_when_missing:
                raise KeyError(
                    "Unable to substitute environment variable placeholder "
                    f"'{key}' because the environment variable is not set."
                )
            else:
                logger.warning(
                    "Unable to substitute environment variable placeholder %s "
                    "because the environment variable is not set, using an "
                    "empty string instead.",
                    key,
                )
                return ""

    def _substitution_func(v: str) -> str:
        return ENV_VARIABLE_PLACEHOLDER_PATTERN.sub(
            _replace_with_env_variable_value, v
        )

    return string_utils.substitute_string(
        value=value, substitution_func=_substitution_func
    )


@contextlib.contextmanager
def temporary_environment(environment: Dict[str, str]) -> Iterator[None]:
    """Temporarily set environment variables.

    Args:
        environment: The environment variables to set.

    Yields:
        Nothing.
    """
    try:
        previous_env = {}
        for key, value in environment.items():
            previous_env[key] = os.environ.get(key, None)
            os.environ[key] = value
        yield
    finally:
        for key, previous_value in previous_env.items():
            updated_value = environment[key]
            current_value = os.environ.get(key)

            if current_value != updated_value:
                # The environment variable got updated while this context
                # manager was active -> We don't reset it back to the old
                # value
                continue
            else:
                if previous_value is None:
                    os.environ.pop(key, None)
                else:
                    os.environ[key] = previous_value


def get_step_environment(
    step_config: "StepConfiguration", stack: "Stack"
) -> Dict[str, str]:
    """Get the environment variables for a step.

    Args:
        step_config: The step configuration.
        stack: The stack on which the step will run.

    Returns:
        A dictionary of environment variables.
    """
    environment = {}
    for component in stack.components.values():
        environment.update(component.environment)

    environment.update(stack.environment)
    environment.update(step_config.environment)

    for key, value in environment.items():
        environment[key] = str(value)

    return environment


def get_step_secret_environment(
    step_config: "StepConfiguration", stack: "Stack"
) -> Dict[str, str]:
    """Get the environment variables for a step.

    Args:
        step_config: The step configuration.
        stack: The stack on which the step will run.

    Returns:
        A dictionary of environment variables.
    """
    # The step secrets contain the pipeline secrets first, followed by the
    # actual secrets defined on the step. We reverse the list to make sure the
    # step secrets override the pipeline secrets.
    secrets = list(reversed(step_config.secrets))
    secrets.extend(stack.secrets)

    for component in stack.components.values():
        secrets.extend(component.secrets)

    # Removes duplicates while preserving order, only the first occurrence of
    # each secret will be kept. We then reverse the list to make sure the
    # overrides are applied in the correct order.
    secrets = list(reversed(dict.fromkeys(secrets)))

    environment = {}
    for secret_name_or_id in secrets:
        try:
            secret = Client().get_secret(secret_name_or_id)
        except Exception as e:
            logger.warning(
                "Failed to get secret `%s` with error: %s. Skipping setting "
                "environment variables for this secret.",
                secret_name_or_id,
                e,
            )
            continue

        if not secret.secret_values:
            logger.warning(
                "Did not find any secret values for secret `%s`. This might be "
                "because you do not have permissions to read the secret "
                "values. Skipping setting environment variables for this "
                "secret.",
                secret_name_or_id,
            )
            continue

        for key, value in secret.secret_values.items():
            environment[key] = str(value)

    return environment



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Pipeline execution utilities."""

import contextvars
from contextlib import contextmanager
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generator,
    Optional,
    Union,
)

from zenml.client import Client
from zenml.config.step_configurations import StepConfigurationUpdate
from zenml.exceptions import RunMonitoringError
from zenml.logger import get_logger
from zenml.models import (
    PipelineRunResponse,
    PipelineSnapshotResponse,
)
from zenml.orchestrators.publish_utils import publish_failed_pipeline_run
from zenml.stack import Stack

if TYPE_CHECKING:
    StepConfigurationUpdateOrDict = Union[
        Dict[str, Any], StepConfigurationUpdate
    ]

logger = get_logger(__name__)


_prevent_pipeline_execution = contextvars.ContextVar(
    "prevent_pipeline_execution", default=False
)


def should_prevent_pipeline_execution() -> bool:
    """Whether to prevent pipeline execution.

    Returns:
        Whether to prevent pipeline execution.
    """
    return _prevent_pipeline_execution.get()


@contextmanager
def prevent_pipeline_execution() -> Generator[None, None, None]:
    """Context manager to prevent pipeline execution.

    Yields:
        None.
    """
    token = _prevent_pipeline_execution.set(True)
    try:
        yield
    finally:
        _prevent_pipeline_execution.reset(token)


def submit_pipeline(
    snapshot: "PipelineSnapshotResponse",
    stack: "Stack",
    placeholder_run: Optional["PipelineRunResponse"] = None,
) -> None:
    """Submit a snapshot for execution.

    Args:
        snapshot: The snapshot to submit.
        stack: The stack on which to submit the snapshot.
        placeholder_run: An optional placeholder run for the snapshot.

    # noqa: DAR401
    Raises:
        BaseException: Any exception that happened while submitting or running
            (in case it happens synchronously) the pipeline.
    """
    # Prevent execution of nested pipelines which might lead to
    # unexpected behavior
    with prevent_pipeline_execution():
        try:
            stack.prepare_pipeline_submission(snapshot=snapshot)
            stack.submit_pipeline(
                snapshot=snapshot,
                placeholder_run=placeholder_run,
            )
        except RunMonitoringError as e:
            # Don't mark the run as failed if the error happened during
            # monitoring of the run.
            raise e.original_exception from None
        except BaseException as e:
            if (
                placeholder_run
                and not Client()
                .get_pipeline_run(placeholder_run.id, hydrate=False)
                .status.is_finished
            ):
                # We failed during/before the submission of the run, so we mark
                # the run as failed if it's still in an unfinished state.
                publish_failed_pipeline_run(placeholder_run.id)

            raise e



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Pipeline configuration classes."""

from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Mapping,
    Optional,
    Tuple,
    Union,
)
from uuid import UUID

from pydantic import (
    ConfigDict,
    Field,
    SerializeAsAny,
    field_validator,
    model_validator,
)

from zenml.artifacts.artifact_config import ArtifactConfig
from zenml.artifacts.external_artifact_config import (
    ExternalArtifactConfiguration,
)
from zenml.client_lazy_loader import ClientLazyLoader
from zenml.config.base_settings import BaseSettings, SettingsOrDict
from zenml.config.cache_policy import CachePolicy, CachePolicyWithValidator
from zenml.config.constants import DOCKER_SETTINGS_KEY, RESOURCE_SETTINGS_KEY
from zenml.config.frozen_base_model import FrozenBaseModel
from zenml.config.retry_config import StepRetryConfig
from zenml.config.source import Source, SourceWithValidator
from zenml.enums import StepRuntime
from zenml.logger import get_logger
from zenml.model.lazy_load import ModelVersionDataLazyLoader
from zenml.model.model import Model
from zenml.utils import deprecation_utils
from zenml.utils.pydantic_utils import before_validator_handler, update_model

if TYPE_CHECKING:
    from zenml.config import DockerSettings, ResourceSettings
    from zenml.config.pipeline_configurations import PipelineConfiguration

logger = get_logger(__name__)


class PartialArtifactConfiguration(FrozenBaseModel):
    """Class representing a partial input/output artifact configuration."""

    materializer_source: Optional[Tuple[SourceWithValidator, ...]] = None
    # TODO: This could be moved to the `PipelineSnapshot` as it's the same
    # for all steps/outputs
    default_materializer_source: Optional[SourceWithValidator] = None
    artifact_config: Optional[ArtifactConfig] = None

    @model_validator(mode="before")
    @classmethod
    @before_validator_handler
    def _remove_deprecated_attributes(
        cls, data: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Removes deprecated attributes from the values dict.

        Args:
            data: The values dict used to instantiate the model.

        Returns:
            The values dict without deprecated attributes.
        """
        deprecated_attributes = ["artifact_source"]

        for deprecated_attribute in deprecated_attributes:
            if deprecated_attribute in data:
                data.pop(deprecated_attribute)

        return data

    @field_validator("materializer_source", mode="before")
    @classmethod
    def _convert_source(
        cls,
        value: Union[None, Source, Dict[str, Any], str, Tuple[Source, ...]],
    ) -> Optional[Tuple[Source, ...]]:
        """Converts old source strings to tuples of source objects.

        Args:
            value: Source string or object.

        Returns:
            The converted source.
        """
        if isinstance(value, str):
            value = (Source.from_import_path(value),)
        elif isinstance(value, dict):
            value = (Source.model_validate(value),)
        elif isinstance(value, Source):
            value = (value,)

        return value


class ArtifactConfiguration(PartialArtifactConfiguration):
    """Class representing a complete input/output artifact configuration."""

    materializer_source: Tuple[SourceWithValidator, ...]

    @field_validator("materializer_source", mode="before")
    @classmethod
    def _convert_source(
        cls,
        value: Union[None, Source, Dict[str, Any], str, Tuple[Source, ...]],
    ) -> Optional[Tuple[Source, ...]]:
        """Converts old source strings to tuples of source objects.

        Args:
            value: Source string or object.

        Returns:
            The converted source.
        """
        if isinstance(value, str):
            value = (Source.from_import_path(value),)
        elif isinstance(value, dict):
            value = (Source.model_validate(value),)
        elif isinstance(value, Source):
            value = (value,)

        return value


class StepConfigurationUpdate(FrozenBaseModel):
    """Class for step configuration updates."""

    enable_cache: Optional[bool] = Field(
        default=None,
        description="Whether to enable cache for the step.",
    )
    enable_artifact_metadata: Optional[bool] = Field(
        default=None,
        description="Whether to store metadata for the output artifacts of "
        "the step.",
    )
    enable_artifact_visualization: Optional[bool] = Field(
        default=None,
        description="Whether to enable visualizations for the output "
        "artifacts of the step.",
    )
    enable_step_logs: Optional[bool] = Field(
        default=None,
        description="Whether to enable logs for the step.",
    )
    step_operator: Optional[Union[bool, str]] = Field(
        default=None,
        description="The step operator to use for the step.",
    )
    experiment_tracker: Optional[Union[bool, str]] = Field(
        default=None,
        description="The experiment tracker to use for the step.",
    )
    parameters: Optional[Dict[str, Any]] = Field(
        default=None,
        description="Parameters for the step function.",
    )
    settings: Optional[Dict[str, SerializeAsAny[BaseSettings]]] = Field(
        default=None,
        description="Settings for the step.",
    )
    environment: Optional[Dict[str, str]] = Field(
        default=None,
        description="The environment for the step.",
    )
    secrets: Optional[List[Union[str, UUID]]] = Field(
        default=None,
        description="The secrets for the step.",
    )
    extra: Optional[Dict[str, Any]] = Field(
        default=None,
        description="Extra configurations for the step.",
    )
    failure_hook_source: Optional[SourceWithValidator] = Field(
        default=None,
        description="The failure hook source for the step.",
    )
    success_hook_source: Optional[SourceWithValidator] = Field(
        default=None,
        description="The success hook source for the step.",
    )
    model: Optional[Model] = Field(
        default=None,
        description="The model to use for the step.",
    )
    retry: Optional[StepRetryConfig] = Field(
        default=None,
        description="The retry configuration for the step.",
    )
    substitutions: Optional[Dict[str, str]] = Field(
        default=None,
        description="The substitutions for the step.",
    )
    cache_policy: Optional[CachePolicyWithValidator] = Field(
        default=None,
        description="The cache policy for the step.",
    )
    runtime: Optional[StepRuntime] = Field(
        default=None,
        description="The step runtime. If not configured, the step will "
        "run inline unless a step operator or docker/resource settings "
        "are configured. This is only applicable for dynamic pipelines.",
    )

    outputs: Mapping[str, PartialArtifactConfiguration] = {}

    def uses_step_operator(self, name: str) -> bool:
        """Checks if the step configuration uses the given step operator.

        Args:
            name: The name of the step operator.

        Returns:
            If the step configuration uses the given step operator.
        """
        if self.step_operator is True:
            return True
        elif isinstance(self.step_operator, str):
            return self.step_operator == name
        else:
            return False

    def uses_experiment_tracker(self, name: str) -> bool:
        """Checks if the step configuration uses the given experiment tracker.

        Args:
            name: The name of the experiment tracker.

        Returns:
            If the step configuration uses the given experiment tracker.
        """
        if self.experiment_tracker is True:
            return True
        elif isinstance(self.experiment_tracker, str):
            return self.experiment_tracker == name
        else:
            return False


class PartialStepConfiguration(StepConfigurationUpdate):
    """Class representing a partial step configuration."""

    name: str
    # TODO: maybe move to spec?
    template: Optional[str] = None
    parameters: Dict[str, Any] = {}
    settings: Dict[str, SerializeAsAny[BaseSettings]] = {}
    environment: Dict[str, str] = {}
    secrets: List[Union[str, UUID]] = []
    extra: Dict[str, Any] = {}
    substitutions: Dict[str, str] = {}
    caching_parameters: Mapping[str, Any] = {}
    external_input_artifacts: Mapping[str, ExternalArtifactConfiguration] = {}
    model_artifacts_or_metadata: Mapping[str, ModelVersionDataLazyLoader] = {}
    client_lazy_loaders: Mapping[str, ClientLazyLoader] = {}
    outputs: Mapping[str, PartialArtifactConfiguration] = {}
    cache_policy: CachePolicyWithValidator = CachePolicy.default()

    # TODO: In Pydantic v2, the `model_` is a protected namespaces for all
    #  fields defined under base models. If not handled, this raises a warning.
    #  It is possible to suppress this warning message with the following
    #  configuration, however the ultimate solution is to rename these fields.
    #  Even though they do not cause any problems right now, if we are not
    #  careful we might overwrite some fields protected by pydantic.
    model_config = ConfigDict(protected_namespaces=())

    # Override the deprecation validator as we do not want to deprecate the
    # `name`` attribute on this class.
    _deprecation_validator = deprecation_utils.deprecate_pydantic_attributes()

    @model_validator(mode="before")
    @classmethod
    @before_validator_handler
    def _remove_deprecated_attributes(cls, data: Any) -> Any:
        """Removes deprecated attributes from the values dict.

        Args:
            data: The values dict used to instantiate the model.

        Returns:
            The values dict without deprecated attributes.
        """
        deprecated_attributes = ["docstring", "inputs"]

        for deprecated_attribute in deprecated_attributes:
            if deprecated_attribute in data:
                data.pop(deprecated_attribute)

        return data


class StepConfiguration(PartialStepConfiguration):
    """Step configuration class."""

    outputs: Mapping[str, ArtifactConfiguration] = {}

    @property
    def resource_settings(self) -> "ResourceSettings":
        """Resource settings of this step configuration.

        Returns:
            The resource settings of this step configuration.
        """
        from zenml.config import ResourceSettings

        model_or_dict: SettingsOrDict = self.settings.get(
            RESOURCE_SETTINGS_KEY, {}
        )

        if isinstance(model_or_dict, BaseSettings):
            model_or_dict = model_or_dict.model_dump()
        return ResourceSettings.model_validate(model_or_dict)

    @property
    def docker_settings(self) -> "DockerSettings":
        """Docker settings of this step configuration.

        Returns:
            The Docker settings of this step configuration.
        """
        from zenml.config import DockerSettings

        model_or_dict: SettingsOrDict = self.settings.get(
            DOCKER_SETTINGS_KEY, {}
        )
        if isinstance(model_or_dict, BaseSettings):
            model_or_dict = model_or_dict.model_dump()
        return DockerSettings.model_validate(model_or_dict)

    def apply_pipeline_configuration(
        self, pipeline_configuration: "PipelineConfiguration"
    ) -> "StepConfiguration":
        """Apply the pipeline configuration to this step configuration.

        Args:
            pipeline_configuration: The pipeline configuration to apply.

        Returns:
            The updated step configuration.
        """
        pipeline_values = pipeline_configuration.model_dump(
            include={
                "settings",
                "extra",
                "failure_hook_source",
                "success_hook_source",
                "retry",
                "substitutions",
                "environment",
                "secrets",
                "cache_policy",
            },
            exclude_none=True,
        )
        if pipeline_values:
            original_values = self.model_dump(
                include={
                    "settings",
                    "extra",
                    "failure_hook_source",
                    "success_hook_source",
                    "retry",
                    "substitutions",
                    "environment",
                    "secrets",
                    "cache_policy",
                },
                exclude_none=True,
            )

            original_values["secrets"] = pipeline_values.get(
                "secrets", []
            ) + original_values.get("secrets", [])

            updated_config_dict = {
                **self.model_dump(),
                **pipeline_values,
            }
            updated_config = self.model_validate(updated_config_dict)
            return update_model(updated_config, original_values)
        else:
            return self.model_copy(deep=True)


class InputSpec(FrozenBaseModel):
    """Step input specification."""

    step_name: str
    output_name: str
    chunk_index: Optional[int] = None
    chunk_size: Optional[int] = None


class StepSpec(FrozenBaseModel):
    """Specification of a pipeline."""

    source: SourceWithValidator
    upstream_steps: List[str]
    # TODO: This should be `Dict[str, List[InputSpec]]`, but that would break
    # client-server compatibility. In the next major release, change this and
    # uncomment the code that migrates legacy specs.
    inputs: Dict[str, Union[List[InputSpec], InputSpec]] = {}
    invocation_id: str
    enable_heartbeat: bool = False

    @model_validator(mode="before")
    @classmethod
    @before_validator_handler
    def _migrate_legacy_fields(cls, data: Dict[str, Any]) -> Dict[str, Any]:
        if "invocation_id" not in data:
            data["invocation_id"] = data.pop("pipeline_parameter_name", "")

        # converted_inputs = {}
        # for key, value in data.get("inputs", {}).items():
        #     if isinstance(value, (InputSpec, dict)):
        #         converted_inputs[key] = [value]
        #     else:
        #         converted_inputs[key] = value
        # data["inputs"] = converted_inputs

        return data

    # TODO: Remove this and use the `inputs` property once we change the type
    # of the `inputs` field.
    @property
    def inputs_v2(self) -> Dict[str, List[InputSpec]]:
        """Inputs of the step spec in v2 format.

        Returns:
            The inputs of the step spec in v2 format.
        """
        return {
            key: [value] if isinstance(value, InputSpec) else value
            for key, value in self.inputs.items()
        }

    def __eq__(self, other: Any) -> bool:
        """Returns whether the other object is referring to the same step.

        This is the case if the other objects is a `StepSpec` with the same
        `upstream_steps` and a `source` that meets one of the following
        conditions:
            - it is the same as the `source` of this step
            - it refers to the same absolute path as the `source` of this step

        Args:
            other: The other object to compare to.

        Returns:
            True if the other object is referring to the same step.
        """
        if isinstance(other, StepSpec):
            if self.upstream_steps != other.upstream_steps:
                return False

            if self.inputs_v2 != other.inputs_v2:
                return False

            if self.invocation_id != other.invocation_id:
                return False

            return self.source.import_path == other.source.import_path

        return NotImplemented


class Step(FrozenBaseModel):
    """Class representing a ZenML step."""

    spec: StepSpec
    config: StepConfiguration
    step_config_overrides: StepConfiguration

    @model_validator(mode="before")
    @classmethod
    @before_validator_handler
    def _add_step_config_overrides_if_missing(cls, data: Any) -> Any:
        """Add step config overrides if missing.

        This is to ensure backwards compatibility with data stored in the DB
        before the `step_config_overrides` field was added. In that case, only
        the `config` field, which contains the merged pipeline and step configs,
        existed. We have no way to figure out which of those values were defined
        on the step vs the pipeline level, so we just use the entire `config`
        object as the `step_config_overrides`.

        Args:
            data: The values dict used to instantiate the model.

        Returns:
            The values dict with the step config overrides added if missing.
        """
        if "step_config_overrides" not in data:
            data["step_config_overrides"] = data["config"]

        return data

    @classmethod
    def from_dict(
        cls,
        data: Dict[str, Any],
        pipeline_configuration: "PipelineConfiguration",
    ) -> "Step":
        """Create a step from a dictionary.

        This method can create a step from data stored without the merged
        `config` attribute, by merging the `step_config_overrides` with the
        pipeline configuration.

        Args:
            data: The dictionary to create the `Step` object from.
            pipeline_configuration: The pipeline configuration to apply to the
                step configuration.

        Returns:
            The instantiated object.
        """
        if "config" not in data:
            config = StepConfiguration.model_validate(
                data["step_config_overrides"]
            )
            data["config"] = config.apply_pipeline_configuration(
                pipeline_configuration
            )
        else:
            # We still need to apply the pipeline substitutions for legacy step
            # objects which include the full config object.
            from zenml.config.pipeline_configurations import (
                PipelineConfiguration,
            )

            config = StepConfiguration.model_validate(data["config"])
            data["config"] = config.apply_pipeline_configuration(
                PipelineConfiguration(
                    name=pipeline_configuration.name,
                    substitutions=pipeline_configuration.substitutions,
                )
            )

        return cls.model_validate(data)



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Endpoint definitions for authentication (login)."""

from typing import List, Optional

from fastapi import APIRouter, Security

import zenml
from zenml.constants import (
    ACTIVATE,
    API,
    INFO,
    LOAD_INFO,
    ONBOARDING_STATE,
    SERVER_SETTINGS,
    STATISTICS,
    VERSION_1,
)
from zenml.enums import AuthScheme
from zenml.exceptions import IllegalOperationError
from zenml.models import (
    ComponentFilter,
    ProjectFilter,
    ServerActivationRequest,
    ServerLoadInfo,
    ServerModel,
    ServerSettingsResponse,
    ServerSettingsUpdate,
    ServerStatistics,
    StackFilter,
    UserResponse,
)
from zenml.zen_server.auth import AuthContext, authorize
from zenml.zen_server.exceptions import error_response
from zenml.zen_server.rbac.models import ResourceType
from zenml.zen_server.rbac.utils import get_allowed_resource_ids
from zenml.zen_server.utils import (
    async_fastapi_endpoint_wrapper,
    server_config,
    zen_store,
)

router = APIRouter(
    prefix=API + VERSION_1,
    tags=["server"],
    responses={401: error_response},
)


@router.get("/version")
async def version() -> str:
    """Get version of the server.

    Returns:
        String representing the version of the server.
    """
    return zenml.__version__


@router.get(
    INFO,
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper
def server_info() -> ServerModel:
    """Get information about the server.

    Returns:
        Information about the server.
    """
    return zen_store().get_store_info()


@router.get(
    LOAD_INFO,
    response_model=ServerLoadInfo,
)
@async_fastapi_endpoint_wrapper
def server_load_info(_: AuthContext = Security(authorize)) -> ServerLoadInfo:
    """Get information about the server load.

    Returns:
        Information about the server load.
    """
    import threading

    # Get the current number of threads
    num_threads = len(threading.enumerate())

    store = zen_store()

    if store.config.driver == "sqlite":
        # SQLite doesn't have a connection pool
        return ServerLoadInfo(
            threads=num_threads,
            db_connections_total=0,
            db_connections_active=0,
            db_connections_overflow=0,
        )

    from sqlalchemy.pool import QueuePool

    # Get the number of connections
    pool = store.engine.pool
    assert isinstance(pool, QueuePool)
    idle_conn = pool.checkedin()
    active_conn = pool.checkedout()
    overflow_conn = max(0, pool.overflow())
    total_conn = idle_conn + active_conn

    return ServerLoadInfo(
        threads=num_threads,
        db_connections_total=total_conn,
        db_connections_active=active_conn,
        db_connections_overflow=overflow_conn,
    )


@router.get(
    ONBOARDING_STATE,
    responses={
        401: error_response,
        404: error_response,
        422: error_response,
    },
)
@async_fastapi_endpoint_wrapper
def get_onboarding_state(
    _: AuthContext = Security(authorize),
) -> List[str]:
    """Get the onboarding state of the server.

    Returns:
        The onboarding state of the server.
    """
    return zen_store().get_onboarding_state()


# We don't have any concrete value that tells us whether a server is a cloud
# workspace, so we use `external_server_id` as the best proxy option.
# For cloud workspaces, we don't add these endpoints as the server settings don't
# have any effect and even allow users to disable functionality that is
# necessary for the cloud onboarding to work.
if server_config().external_server_id is None:

    @router.get(
        SERVER_SETTINGS,
        responses={
            401: error_response,
            404: error_response,
            422: error_response,
        },
    )
    @async_fastapi_endpoint_wrapper
    def get_settings(
        _: AuthContext = Security(authorize),
        hydrate: bool = True,
    ) -> ServerSettingsResponse:
        """Get settings of the server.

        Args:
            hydrate: Whether to hydrate the response.

        Returns:
            Settings of the server.
        """
        return zen_store().get_server_settings(hydrate=hydrate)

    @router.put(
        SERVER_SETTINGS,
        responses={
            401: error_response,
            404: error_response,
            422: error_response,
        },
    )
    @async_fastapi_endpoint_wrapper
    def update_server_settings(
        settings_update: ServerSettingsUpdate,
        auth_context: AuthContext = Security(authorize),
    ) -> ServerSettingsResponse:
        """Updates the settings of the server.

        Args:
            settings_update: Settings update.
            auth_context: Authentication context.

        Raises:
            IllegalOperationError: If trying to update admin properties without
                admin permissions.

        Returns:
            The updated settings.
        """
        if not server_config().rbac_enabled:
            will_update_admin_properties = bool(
                settings_update.model_dump(
                    exclude_none=True, exclude={"onboarding_state"}
                )
            )

            if not auth_context.user.is_admin and will_update_admin_properties:
                raise IllegalOperationError(
                    "Only admins can update server settings."
                )

        return zen_store().update_server_settings(settings_update)


# When the auth scheme is set to EXTERNAL, users cannot be managed via the
# API and the server is activated on deployment
if server_config().auth_scheme != AuthScheme.EXTERNAL:

    @router.put(
        ACTIVATE,
        responses={
            401: error_response,
            404: error_response,
            422: error_response,
        },
    )
    @async_fastapi_endpoint_wrapper
    def activate_server(
        activate_request: ServerActivationRequest,
    ) -> Optional[UserResponse]:
        """Updates a stack.

        Args:
            activate_request: The request to activate the server.

        Returns:
            The default admin user that was created during activation, if any.
        """
        return zen_store().activate_server(activate_request)


@router.get(
    STATISTICS,
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper
def get_server_statistics(
    auth_context: AuthContext = Security(authorize),
) -> ServerStatistics:
    """Gets server statistics.

    Args:
        auth_context: Authentication context.

    Returns:
        Statistics of the server.
    """
    user_id = auth_context.user.id
    component_filter = ComponentFilter()
    component_filter.configure_rbac(
        authenticated_user_id=user_id,
        id=get_allowed_resource_ids(
            resource_type=ResourceType.STACK_COMPONENT
        ),
    )

    project_filter = ProjectFilter()
    project_filter.configure_rbac(
        authenticated_user_id=user_id,
        id=get_allowed_resource_ids(resource_type=ResourceType.PROJECT),
    )

    stack_filter = StackFilter()
    stack_filter.configure_rbac(
        authenticated_user_id=user_id,
        id=get_allowed_resource_ids(resource_type=ResourceType.STACK),
    )

    return ServerStatistics(
        stacks=zen_store().count_stacks(filter_model=stack_filter),
        components=zen_store().count_stack_components(
            filter_model=component_filter
        ),
        projects=zen_store().count_projects(filter_model=project_filter),
    )
